"""
Tools for binning data.
"""

import numpy
import numpy as np
import collections

############################################################

def centers(array):
    """
    Centers corresponding to bin edges (assuming regular linear intervals).
    """
    return array[0: -1] + (0.5 * (array[1] - array[0]))

############################################################

def take2D(histogram, x, y, bins_x, bins_y):
    """
    Take the value from a two-dimensional histogram from the bin corresponding to (x, y).
    """
    histogram = numpy.array(histogram)
    
    if numpy.isscalar(x):
        x = [x]
    if numpy.isscalar(y):
        y = [y]

    bins_x[-1] += 1.e-10 * (bins_x[-1] - bins_x[-2]) # Numerical stability
    bins_y[-1] += 1.e-10 * (bins_y[-1] - bins_y[-2])

    #return numpy.take(histogram, (histogram.shape[1] * (numpy.digitize(y, bins_y) - 1)) + (numpy.digitize(x, bins_x) - 1))

    # Return numpy.nan for entries which are outside the binning range on either axis
    index = (histogram.shape[1] * (numpy.digitize(y, bins_y) - 1)) + (numpy.digitize(x, bins_x) - 1)
    index_clipped = numpy.clip(index, 0, (histogram.shape[0] * histogram.shape[1]) - 1)
    val = numpy.take(histogram, index_clipped)

    outlier_x = numpy.logical_or(x < bins_x[0], x > bins_x[-1])
    outlier_y = numpy.logical_or(y < bins_y[0], y > bins_y[-1])
    outlier = numpy.logical_or(outlier_x, outlier_y)
    val[outlier] = numpy.nan

    return val 
    
############################################################

def cloudInCells(x, y, bins, weights=None):
    """
    Use cloud-in-cells binning algorithm. Only valid for equal-spaced linear bins.
    http://ta.twi.tudelft.nl/dv/users/Lemmens/MThesis.TTH/chapter4.html#tth_sEc2
    http://www.gnu.org/software/archimedes/manual/html/node29.html

    INPUTS:
        x: array of x-values
        y: array or y-values
        bins: [bins_x, bins_y] format, where bins_x corresponds to the bin edges along x-axis
        weights[None]: optionally assign a weight to each entry
    OUTPUTS:
        histogram:
        bins_x:
        bins_y:
    """

    # For consistency, the variable names should be changed in this function, but low priority...
    
    x_bins = numpy.array(bins[0])
    delta_x = x_bins[1] - x_bins[0]
    # Overflow and underflow bins
    x_bins = numpy.insert(x_bins, 0, x_bins[0] - delta_x)
    x_bins = numpy.append(x_bins, x_bins[-1] + delta_x)
    y_bins = numpy.array(bins[1])
    delta_y = y_bins[1] - y_bins[0]
    y_bins = numpy.insert(y_bins, 0, y_bins[0] - delta_y)
    y_bins = numpy.append(y_bins, y_bins[-1] + delta_y)

    x_bound_cut = numpy.logical_and(x >= x_bins[0], x <= x_bins[-1])
    y_bound_cut = numpy.logical_and(y >= y_bins[0], y <= y_bins[-1])
    bound_cut = numpy.logical_and(x_bound_cut, y_bound_cut)

    if not numpy.any(weights):
        bound_weights = numpy.ones(len(x))[bound_cut]
    else:
        bound_weights = numpy.array(weights)[bound_cut]

    x_vals = numpy.array(x)[bound_cut]
    y_vals = numpy.array(y)[bound_cut]

    x_width = x_bins[1] - x_bins[0]
    y_width = y_bins[1] - y_bins[0]

    x_centers = x_bins[0: -1] + (0.5 * x_width)  
    y_centers = y_bins[0: -1] + (0.5 * y_width)

    dx = x_vals - x_centers[numpy.digitize(x_vals, x_bins) - 1]
    dy = y_vals - y_centers[numpy.digitize(y_vals, y_bins) - 1]

    ux = ((dx / x_width) * (dx >= 0)) +\
         ((1. + (dx / x_width)) * (dx < 0))
    lx = 1. - ux

    uy = ((dy / y_width) * (dy >= 0)) +\
         ((1. + (dy / y_width)) * (dy < 0))
    ly = 1. - uy

    new_x_vals = []
    new_y_vals = []
    cell_weights = []

    # 4 corners
    new_x_vals.append(x_vals + (0.5 * x_width))
    new_y_vals.append(y_vals + (0.5 * y_width))
    cell_weights.append(bound_weights * ux * uy)

    new_x_vals.append(x_vals + (0.5 * x_width))
    new_y_vals.append(y_vals - (0.5 * y_width))
    cell_weights.append(bound_weights * ux * ly)
    
    new_x_vals.append(x_vals - (0.5 * x_width))
    new_y_vals.append(y_vals + (0.5 * y_width))
    cell_weights.append(bound_weights * lx * uy)
    
    new_x_vals.append(x_vals - (0.5 * x_width))
    new_y_vals.append(y_vals - (0.5 * y_width))
    cell_weights.append(bound_weights * lx * ly)

    new_x_vals = numpy.concatenate(new_x_vals)
    new_y_vals = numpy.concatenate(new_y_vals)
    cell_weights = numpy.concatenate(cell_weights)

    result = numpy.histogram2d(new_x_vals, new_y_vals,
                               bins = [x_bins, y_bins],
                               weights = cell_weights)[0]
 
    result = numpy.transpose(result[1: result.shape[0] - 1])[1: result.shape[1] - 1]

    return result, x_bins, y_bins


def kernelDensity(x, y, bins, weights=None, fx=1, fy=1):
    nx,ny = (len(bins[0])-1)*fx,(len(bins[1])-1)*fy
    xmin,xmax = bins[0].min(),bins[0].max()
    ymin,ymax = bins[1].min(),bins[1].max()
    # Correct weights for CMD bin size
    if weights is not None:
        weights = np.asarray(weights)
        weights *= 1/float(fx*fy)
    grid = fast_kde(x,y,gridsize=(nx,ny),extents=[xmin,xmax,ymin,ymax])
    results = sum_chunks(grid,fx,fy)
    return results,bins[0],bins[1]
    
############################################################

"""
A faster gaussian kernel density estimate (KDE).
Intended for computing the KDE on a regular grid (different use case than 
scipy's original scipy.stats.kde.gaussian_kde()).
@author: Joe Kington

@note: http://pastebin.com/LNdYCZgw
"""
__license__ = 'MIT License <http://www.opensource.org/licenses/mit-license.php>'

import numpy as np
import scipy as sp
import scipy.sparse
import scipy.signal

def fast_kde(x, y, gridsize=(200,200), extents=None, nocorrelation=False, weights=None):
    """
    Performs a gaussian kernel density estimate over a regular grid using a
    convolution of the gaussian kernel with a 2D histogram of the data.

    This function is typically several orders of magnitude faster than 
    scipy.stats.kde.gaussian_kde for large (>1e7) numbers of points and 
    produces an essentially identical result.

    Input:
        x: The x-coords of the input data points
        y: The y-coords of the input data points
        gridsize: (default: 200x200) A (nx,ny) tuple of the size of the output 
            grid
        extents: (default: extent of input data) A (xmin, xmax, ymin, ymax)
            tuple of the extents of output grid
        nocorrelation: (default: False) If True, the correlation between the
            x and y coords will be ignored when preforming the KDE.
        weights: (default: None) An array of the same shape as x & y that 
            weighs each sample (x_i, y_i) by each value in weights (w_i).
            Defaults to an array of ones the same size as x & y.
    Output:
        A gridded 2D kernel density estimate of the input points. 
    """
    #---- Setup --------------------------------------------------------------
    x, y = np.asarray(x), np.asarray(y)
    x, y = np.squeeze(x), np.squeeze(y)
    
    if x.size != y.size:
        raise ValueError('Input x & y arrays must be the same size!')

    nx, ny = gridsize
    n = x.size

    if weights is None:
        # Default: Weight all points equally
        weights = np.ones(n)
    else:
        weights = np.squeeze(np.asarray(weights))
        if weights.size != x.size:
            raise ValueError('Input weights must be an array of the same size'
                    ' as input x & y arrays!')

    # Default extents are the extent of the data
    if extents is None:
        xmin, xmax = x.min(), x.max()
        ymin, ymax = y.min(), y.max()
    else:
        xmin, xmax, ymin, ymax = map(float, extents)
    dx = (xmax - xmin) / (nx - 1)
    dy = (ymax - ymin) / (ny - 1)

    #---- Preliminary Calculations -------------------------------------------

    # First convert x & y over to pixel coordinates
    # (Avoiding np.digitize due to excessive memory usage in numpy < v1.5!)
    # http://stackoverflow.com/q/8805601/
    xyi = np.vstack((x,y)).T
    xyi -= [xmin, ymin]
    xyi /= [dx, dy]
    xyi = np.floor(xyi, xyi).T

    # Next, make a 2D histogram of x & y
    # Avoiding np.histogram2d due to excessive memory usage with many points
    # http://stackoverflow.com/q/8805601/
    grid = sp.sparse.coo_matrix((weights, xyi), shape=(nx, ny)).toarray()

    # Calculate the covariance matrix (in pixel coords)
    cov = np.cov(xyi)

    if nocorrelation:
        cov[1,0] = 0
        cov[0,1] = 0

    # Scaling factor for bandwidth
    scotts_factor = np.power(n, -1.0 / 6) # For 2D

    #---- Make the gaussian kernel -------------------------------------------

    # First, determine how big the kernel needs to be
    std_devs = np.diag(np.sqrt(cov))
    kern_nx, kern_ny = np.round(scotts_factor * 2 * np.pi * std_devs)

    # Determine the bandwidth to use for the gaussian kernel
    inv_cov = np.linalg.inv(cov * scotts_factor**2) 

    # x & y (pixel) coords of the kernel grid, with <x,y> = <0,0> in center
    xx = np.arange(kern_nx, dtype=np.float) - kern_nx / 2.0
    yy = np.arange(kern_ny, dtype=np.float) - kern_ny / 2.0
    xx, yy = np.meshgrid(xx, yy)

    # Then evaluate the gaussian function on the kernel grid
    kernel = np.vstack((xx.flatten(), yy.flatten()))
    kernel = np.dot(inv_cov, kernel) * kernel 
    kernel = np.sum(kernel, axis=0) / 2.0 
    kernel = np.exp(-kernel) 
    kernel = kernel.reshape((kern_ny, kern_nx))

    #---- Produce the kernel density estimate --------------------------------

    # Convolve the gaussian kernel with the 2D histogram, producing a gaussian
    # kernel density estimate on a regular grid
    grid = sp.signal.convolve2d(grid, kernel, mode='same', boundary='fill').T

    ### ADW: Commented out for 
    ### # Normalization factor to divide result by so that units are in the same
    ### # units as scipy.stats.kde.gaussian_kde's output.  
    ### norm_factor = 2 * np.pi * cov * scotts_factor**2
    ### norm_factor = np.linalg.det(norm_factor)
    ### norm_factor = n * dx * dy * np.sqrt(norm_factor)
    ###  
    ### # Normalize the result
    ### grid /= norm_factor

    return grid

def sum_chunks(data,fx=4,fy=4):
    y,x = data.shape
    binned = data.reshape(y//fy, fy, x//fx, fx)
    return binned.sum(axis=3).sum(axis=1)

############################################################

def reverseHistogram(data,bins=None):
    """             
    Bins data using numpy.histogram and calculates the
    reverse indices for the entries like IDL.
    Parameters:
    data  : data to pass to numpy.histogram
    bins  : bins to pass to numpy.histogram 
    Returns: 
    hist  : bin content output by numpy.histogram 
    edges : edges output from numpy.histogram 
    rev   : reverse indices of entries in each bin 
    Using Reverse Indices: 
        h,e,rev = histogram(data, bins=bins) 
        for i in range(h.size):  
            if rev[i] != rev[i+1]: 
                # data points were found in this bin, get their indices
                indices = rev[ rev[i]:rev[i+1] ] 
                # do calculations with data[indices] ...  
    """
    if bins is None: bins = numpy.arange(data.max()+2)
    hist, edges = numpy.histogram(data, bins=bins)
    digi = numpy.digitize(data.flat,bins=numpy.unique(data)).argsort()
    rev = numpy.hstack( (len(edges), len(edges) + numpy.cumsum(hist), digi) )
    return hist,edges,rev
    
def binnedMedian(x,y,xbins=None):
    hist,edges,rev = reverseHistogram(x,bins=xbins)
    avg_x = (edges[:-1]+edges[1:])/2.
    med_y = []
    for i,n in enumerate(hist):
        indices = rev[ rev[i]:rev[i+1] ] 
        med_y.append( numpy.median(y[indices]) )
    return avg_x, numpy.array(med_y)



#!/usr/bin/env python
import os, sys
from os.path import splitext, exists
import subprocess

def fits2ply(infile,outfile=None):
    base,ext = os.path.splitext(infile)
    if ext != '.fits': 
        raise Exception("Infile must be FITS.")

    if outfile is None:
        outfile = base + '.ply'
    if os.path.exists(outfile): os.remove(outfile)

    print "Converting %s to %s"%(infile,outfile)

    ### Python Routine
    import mangle
    mask = mangle.Mangle(infile)
    mask.write(outfile)

    #### IDL Routine
    ## Should use a tmp file...
    #profile = 'fits2ply.pro'
    #if os.path.exists(profile): os.remove(profile)
    #pro = open(profile,'w')
    #pro.write("""
    #pro fits2ply
    #  args = command_line_args()
    #  infile = args[0]
    #  outfile = args[1]
    # 
    #  read_fits_polygons, infile, polygons
    #  write_mangle_polygons, outfile, polygons
    #end"""
    #)
    #pro.close()
    # 
    #cmd = 'idl -e "fits2ply" -args %s %s'%(infile, outfile)
    #print cmd
    #subprocess.call(cmd, shell=True)
    #os.remove(profile)

def poly2poly(infile, outfile, *opts):
    opts = ' '.join(opts)
    cmd = "poly2poly %s %s %s"%(opts,infile,outfile)
    print cmd
    subprocess.call(cmd,shell=True)

def pixelize(infile, outfile, *opts):
    opts = ' '.join(opts)
    cmd = "pixelize %s %s %s"%(opts,infile,outfile)
    print cmd
    subprocess.call(cmd,shell=True)

def snap(infile, outfile, *opts):
    opts = ' '.join(opts)
    cmd = "snap %s %s %s"%(opts,infile,outfile)
    print cmd
    subprocess.call(cmd,shell=True)
    
def balkanize(infile, outfile, *opts):
    opts = ' '.join(opts)
    cmd = "balkanize %s %s %s"%(opts,infile,outfile)
    print cmd
    subprocess.call(cmd,shell=True)
        
def unify(infile, outfile, *opts):
    opts = ' '.join(opts)
    cmd = "unify %s %s %s"%(balkfile,unifile,opts)
    print cmd
    subprocess.call(cmd,shell=True)    

if __name__ == "__main__":
    from optparse import OptionParser
    usage = "Usage: %prog  [options] input"
    description = "python script"
    parser = OptionParser(usage=usage,description=description)
    parser.add_option("-p","--pix",default=8, type='int',
                      help="Mangle pixelization")
    parser.add_option("-s","--snap",default=0.1, type='float',
                      help="Snap tolerance.")
    (opts, args) = parser.parse_args()

    if not len(args):
        print "ERROR: Infile require."
        parser.print_help()
        sys.exit(1)
     
    infile = args[0]
    base, ext = os.path.splitext(infile)
    #windows = ['window.boss.dr9.ply']
    #masks   = ['mask.boss.dr9.field.ply','mask.boss.dr9.star.ply']
    #holes   = []
    #base    = splitext(windows[0])[0]
    ## This is to match up with the HEALPix polygons
    #if   nside > 32: pix = 8
    #elif nside < 16: pix = 3
    #else:            pix = 5
        
    ## Polygonize
    #if ext == ".fits":
    #    polyfile = base+'.ply'
    #    cmd = "fits2ply %s %s"%(infile,polyfile)
    #    print cmd
    #    subprocess.call(cmd,shell=True)
    #else:
    #    polyfile = infile
    #if masks:
    #    jw = open('jw','w')
    #    jw.write('0\n')
    #    jw.close()
    # 
    #for mask in masks:
    #    hole = mask.replace('mask.','holes.')
    #    if not os.path.exists(hole):
    #        cmd = "weight -zjw %s %s"%(mask,hole)
    #        print cmd
    #        subprocess.call(cmd,shell=True)
    #    holes.append(hole)

    # To create the holy file
    # ???
    
    snapopts= "-a.1 -b.1 -t.1"
    pixopts = "-Ps0,%i"%opts.pix
    mtol="-m1e-8"
    polyopts = '-k1e-11'

    # Poly2Poly
    #polyfile = base + '.poly.ply'
    #if not os.path.exists(polyfile):
    #    poly2poly(infile,polyfile,polyopts,mtol)
    #else: print "Found %s; skipping..."%polyfile
    # 
    ## Pre-snap (only self-snap)
    #snapfile =  base + '.psnap.ply'
    #if not os.path.exists(snapfile):
    #    snap(infile,snapfile,"-S ",snapopts,mtol)
    #else: print "Found %s; skipping..."%snapfile
    # 
    ## Pixelize
    pixfile = base + '.pix.ply'
    if not os.path.exists(pixfile):
        pixelize(infile,pixfile,pixopts,mtol)
    else: print "Found %s; skipping..."%pixfile

    # Snap
    snapfile =  base + '.snap.ply'
    if not os.path.exists(snapfile):
        snap(pixfile,snapfile,snapopts,mtol)
    else: print "Found %s; skipping..."%snapfile

    # Balkanize
    balkfile = base + '.balk.ply'
    if not os.path.exists(balkfile):
        balkanize(snapfile,balkfile,mtol)
    else: print "Found %s; skipping..."%balkfile

    # Unify
    unifile = base + '.uni.ply'
    if not os.path.exists(unifile):
        unify(balkfile, unifile)
    else: print "Found %s; skipping..."%unifile




"""
Define a region of interest (ROI) in color, magnitude, and direction space.

The ROI is divided into 3 regions:
1) The 'target' region: The region occuping the likelihood-scale healpix
   pixel over which the likelihood is evaluated. (Size controlled by
   'nside_likelihood')
2) The 'interior' region: The region where objects are included into the
   likelihood fit.
3) The 'annulus' region: The region where the background is fit.

"""

import numpy
import numpy as np
import healpy

import ugali.utils.binning
import ugali.utils.projector
import ugali.utils.skymap

from ugali.utils.config import Config
from ugali.utils.healpix import query_disc, ang2pix, pix2ang, ang2vec

############################################################

# ADW: Should really write some "PixelSet" object that contains the pixels for each region...

class PixelRegion(np.ndarray):

    def __new__(cls, nside, pixels):
        # Input array is an already formed ndarray instance
        # We first cast to be our class type
        obj = np.asarray(pixels).view(cls)
        # add the new attribute to the created instance
        obj._nside = nside
        obj._pix = pixels
        obj._lon,obj._lat = pix2ang(nside,pixels)
        # Finally, we must return the newly created object:
        return obj

    def __array_finalize__(self, obj):
        # see InfoArray.__array_finalize__ for comments
        if obj is None: return

    @property
    def lon(self):
        return self._lon

    @property
    def lat(self):
        return self._lat

    @property
    def nside(self):
        return self._nside

    @property
    def pix(self):
        return self._pix

class ROI(object):

    def __init__(self, config, lon, lat):

        self.config = Config(config)
        self.lon = lon
        self.lat = lat

        self.projector = ugali.utils.projector.Projector(self.lon, self.lat)

        self.vec = vec = ang2vec(self.lon, self.lat)
        self.pix = ang2pix(self.config['coords']['nside_likelihood'],self.lon,self.lat)

        # Pixels from the entire ROI disk
        pix = query_disc(self.config['coords']['nside_pixel'], vec, 
                         self.config['coords']['roi_radius'])
        self.pixels = PixelRegion(self.config['coords']['nside_pixel'],pix)

        # Pixels in the interior region
        pix = query_disc(self.config['coords']['nside_pixel'], vec, 
                         self.config['coords']['roi_radius_interior'])
        self.pixels_interior = PixelRegion(self.config['coords']['nside_pixel'],pix)

        # Pixels in the outer annulus
        pix = query_disc(self.config['coords']['nside_pixel'], vec, 
                         self.config['coords']['roi_radius_annulus'])
        pix = numpy.setdiff1d(self.pixels, pix)
        self.pixels_annulus = PixelRegion(self.config['coords']['nside_pixel'],pix)

        # Pixels within target healpix region
        pix = ugali.utils.skymap.subpixel(self.pix,self.config['coords']['nside_likelihood'],
                                          self.config['coords']['nside_pixel'])
        self.pixels_target = PixelRegion(self.config['coords']['nside_pixel'],pix)

        # Boolean arrays for selecting given pixels 
        # (Careful, this works because pixels are pre-sorted by query_disc before in1d)
        self.pixel_interior_cut = numpy.in1d(self.pixels, self.pixels_interior)

        # ADW: Updated for more general ROI shapes
        #self.pixel_annulus_cut  = ~self.pixel_interior_cut
        self.pixel_annulus_cut  = numpy.in1d(self.pixels, self.pixels_annulus)

        # # These should be unnecessary now
        # self.centers_lon, self.centers_lat = self.pixels.lon, self.pixels.lat
        # self.centers_lon_interior,self.centers_lat_interior = self.pixels_interior.lon,self.pixels_interior.lat
        # self.centers_lon_target, self.centers_lat_target = self.pixels_target.lon, self.pixels_target.lat

        self.area_pixel = healpy.nside2pixarea(self.config.params['coords']['nside_pixel'],degrees=True) # deg^2
                                     
        """
        self.centers_x = self._centers(self.bins_x)
        self.centers_y = self._centers(self.bins_y)

        self.delta_x = self.config.params['coords']['pixel_size']
        self.delta_y = self.config.params['coords']['pixel_size']
        
        # Should actually try to take projection effects into account for better accuracy
        # MC integration perhaps?
        # Throw points in a cone around full ROI and see what fraction fall in
        self.area_pixel = self.config.params['coords']['pixel_size']**2
        
        self.centers_lon, self.centers_lat = self.projector.imageToSphere(self.centers_x, self.centers_y)
        """

        # ADW: These are really bin edges, should be careful and consistent
        self.bins_mag = numpy.linspace(self.config.params['mag']['min'],
                                       self.config.params['mag']['max'],
                                       self.config.params['mag']['n_bins'] + 1)
        
        self.bins_color = numpy.linspace(self.config.params['color']['min'],
                                         self.config.params['color']['max'],
                                         self.config.params['color']['n_bins'] + 1)

        self.centers_mag = ugali.utils.binning.centers(self.bins_mag)
        self.centers_color = ugali.utils.binning.centers(self.bins_color)

        self.delta_mag = self.bins_mag[1] - self.bins_mag[0]
        self.delta_color = self.bins_color[1] - self.bins_color[0]

        # Axis labels
        self.label_x = 'x (deg)'
        self.label_y = 'y (deg)'
        
        if self.config.params['catalog']['band_1_detection']:
            self.label_mag = '%s (mag)'%(self.config.params['catalog']['mag_1_field'])
        else:
            self.label_mag = '%s (mag)'%(self.config.params['catalog']['mag_2_field'])
        self.label_color = '%s - %s (mag)'%(self.config.params['catalog']['mag_1_field'],
                                            self.config.params['catalog']['mag_2_field'])

        #self.precomputeAngsep()

    def plot(self, value=None, pixel=None):
        """
        Plot the ROI
        """
        import ugali.utils.plotting

        map_roi = numpy.array(healpy.UNSEEN \
                              * numpy.ones(healpy.nside2npix(self.config.params['coords']['nside_pixel'])))
        
        if value is None:
            #map_roi[self.pixels] = ugali.utils.projector.angsep(self.lon, self.lat, self.centers_lon, self.centers_lat)
            map_roi[self.pixels] = 1
            map_roi[self.pixels_annulus] = 0
            map_roi[self.pixels_target] = 2
        elif value is not None and pixel is None:
            map_roi[self.pixels] = value
        elif value is not None and pixel is not None:
            map_roi[pixel] = value
        else:
            logger.error("Can't parse input")
        
        ugali.utils.plotting.zoomedHealpixMap('Region of Interest',
                                              map_roi,
                                              self.lon, self.lat,
                                              self.config.params['coords']['roi_radius'])

    # ADW: Maybe these should be associated with the PixelRegion objects
    def inPixels(self,lon,lat,pixels):
        """ Function for testing if coordintes in set of ROI pixels. """
        nside = self.config.params['coords']['nside_pixel']
        return ugali.utils.healpix.in_pixels(lon,lat,pixels,nside)
        
    def inROI(self,lon,lat):
        return self.inPixels(lon,lat,self.pixels)

    def inAnnulus(self,lon,lat):
        return self.inPixels(lon,lat,self.pixels_annulus)

    def inInterior(self,lon,lat):
        return self.inPixels(lon,lat,self.pixels_interior)

    def inTarget(self,lon,lat):
        return self.inPixels(lon,lat,self.pixels_target)

    def indexPixels(self,lon,lat,pixels):
        nside = self.config.params['coords']['nside_pixel']
        return ugali.utils.healpix.index_pixels(lon,lat,pixels,nside)

    def indexROI(self,lon,lat):
        return self.indexPixels(lon,lat,self.pixels)

    def indexAnnulus(self,lon,lat):
        return self.indexPixels(lon,lat,self.pixels_annulus)

    def indexInterior(self,lon,lat):
        return self.indexPixels(lon,lat,self.pixels_interior)

    def indexTarget(self,lon,lat):
        return self.indexPixels(lon,lat,self.pixels_target)
        
    def getCatalogPixels(self):
        """
        Return the catalog pixels spanned by this ROI.
        """
        filenames = self.config.getFilenames()

        nside_catalog = self.config.params['coords']['nside_catalog']
        nside_pixel = self.config.params['coords']['nside_pixel']
        # All possible catalog pixels spanned by the ROI
        superpix = ugali.utils.skymap.superpixel(self.pixels,nside_pixel,nside_catalog)
        superpix = numpy.unique(superpix)
        # Only catalog pixels that exist in catalog files
        pixels = numpy.intersect1d(superpix, filenames['pix'].compressed())
        return pixels
        
############################################################




#!/usr/bin/env python
"""
Base functionality for pipeline scripts
"""

import ugali.utils.batch
#from ugali.utils.batch import factory as batchFactory

from ugali.utils.parser import Parser
from ugali.utils.logger import logger
from ugali.utils.config import Config

class Pipeline(object):
    """
    A pipeline script owns:
    - A set of command line arguments
    - A set of runtime components
    """

    def __init__(self, description=__doc__, components=[]):
        self.description = description
        self.components = components
        self._setup_parser()

    def _setup_parser(self):
        self.parser = Parser(description=self.description)
        self.parser.add_config()
        self.parser.add_debug()
        self.parser.add_force()
        self.parser.add_queue()
        self.parser.add_run(choices=self.components) 
        self.parser.add_verbose()
        self.parser.add_version()

    def parse_args(self):
        self.opts = self.parser.parse_args()
        if not self.opts.run: 
            self.opts.run = self.components

        self.config = Config(self.opts.config)        
        self.batch = ugali.utils.batch.batchFactory(self.opts.queue)

    def run(self):
        logger.warning("Doing nothing...")
        return

    def execute(self):
        ret = self.run()
        logger.info("Done.")
        return ret

if __name__ == "__main__":
    description = "Pipeline test"
    components = ['test']
     
    def run(self):
        logger.info("Testing pipeline...")
        if 'test' in self.opts.run:
            logger.info("  This should run.")
        if 'foo' in self.opts.run:
            logger.error("  This should NOT run")
            raise Exception
    Pipeline.run = run
        
    pipeline = Pipeline(description,components)
    pipeline.parser.print_help()
    pipeline.parse_args()
    pipeline.execute()




"""
Class for storing and updating config dictionaries.
"""
import os,sys
from os.path import join, exists
import pprint
import copy
import numpy
import numpy.ma.mrecords
import healpy

from ugali.utils.logger import logger
import ugali.utils.config # To recognize own type

try: import yaml
except ImportError: logger.warning("YAML not found")

class Config(dict):
    """
    Configuration object
    """

    def __init__(self, input, default=None):
        """
        Initialize a configuration object from a filename or a dictionary.
        Provides functionality to merge with a default configuration.

        Parameters:
          input:   Either filename or dictionary (deep copied)
          default: Default configuration to merge
        
        Returns:
          config
        """
        self.update(self._load(default))
        self.update(self._load(input))

        # For back-compatibility...
        self.params = self

        # Possible filenames from this config (masked by existence)
        try:
            self.filenames = self.getFilenames()
            self._makeFilenames()
        except:
            exc_type,exc_value,exc_traceback = sys.exc_info()
            logger.warning("%s %s"%(exc_type,exc_value))
            logger.warning("Filenames could not be created for config.")

    def __str__(self):
        return yaml.dump(self)

    def _load(self, input):
        if isinstance(input, basestring):
            self.filename = input
            ext = os.path.splitext(input)[1]
            if ext == '.py':
                # ADW: This is dangerous and terrible!!!
                # THIS SHOULD BE DEPRICATED!!!
                reader = open(input)
                params = eval(''.join(reader.readlines()))
                reader.close()
            elif ext == '.yaml':
                params = yaml.load(open(input))
            else:
                raise Exception('Unrecognized config format: %s'%ext)
        elif isinstance(input, Config):
            # This is the copy constructor...
            self.filename = input.filename
            params = copy.deepcopy(input)
        elif isinstance(input, dict):
            params = copy.deepcopy(input)
        elif input is None:
            params = {}
        else:
            raise Exception('Unrecognized input')

        return params

    def _makeFilenames(self):
        likedir=self['output']['likedir']
        self.likefile  = join(likedir,self['output']['likefile'])
        self.mergefile = join(likedir,self['output']['mergefile'])
        self.roifile   = join(likedir,self['output']['roifile'])

        searchdir=self['output']['searchdir']
        self.labelfile  = join(searchdir,self['output']['labelfile'])
        self.objectfile = join(searchdir,self['output']['objectfile'])
        self.assocfile  = join(searchdir,self['output']['assocfile'])
        self.candfile   = join(searchdir,self['output']['candfile'])

        mcmcdir=self['output']['mcmcdir']
        self.mcmcfile   = join(mcmcdir,self['output']['mcmcfile'])

    def write(self, outfile):
        ext = os.path.splitext(outfile)[1]
        writer = open(outfile, 'w')
        if ext == '.py':
            writer.write(pprint.pformat(self))
        elif ext == '.yaml':
            writer.write(yaml.dump(self))
        else:
            writer.close()
            raise Exception('Unrecognized config format: %s'%ext)
        writer.close()

    def getFilenames(self,pixels=None):
        """
        Create a masked records array of all filenames for the given set of
        pixels and store the existence of those files in the mask values.
        If pixels is None, default behavior is to try to join grab
        dirname + basename with no pixel insertion. If pixels == -1, grab
        all pixels of 'nside_catalog'.

        Examples:
        f = getFilenames([1,2,3])
        # All possible catalog files
        f['catalog'].data
        # All existing catalog files
        f['catalog'][~f.mask['catalog']]
        # or
        f['catalog'].compressed()
        # All missing mask_1 files
        f['mask_1'][f.mask['mask_1']]
        # Pixels where all files exist
        f['pix'][~f.mask['pix']]
        """
        nside_catalog = self['coords']['nside_catalog']

        if nside_catalog is None:
            pixels = [None]
        elif pixels is not None:
            pixels = [pixels] if numpy.isscalar(pixels) else pixels
        else:
            pixels = numpy.arange(healpy.nside2npix(nside_catalog))   

        npix = len(pixels)

        catalog_dir = self['catalog']['dirname']
        catalog_base = self['catalog']['basename']
         
        mask_dir = self['mask']['dirname']
        mask_base_1 = self['mask']['basename_1']
        mask_base_2 = self['mask']['basename_2']
         
        data = numpy.ma.empty(npix,dtype=[('pix',int), ('catalog',object), 
                                          ('mask_1',object), ('mask_2',object)])
        mask = numpy.ma.empty(npix,dtype=[('pix',bool), ('catalog',bool), 
                                          ('mask_1',bool), ('mask_2',bool)])
        for ii,pix in enumerate(pixels):
            if pix is None:
                catalog = os.path.join(catalog_dir,catalog_base)
                mask_1 = os.path.join(mask_dir,mask_base_1)
                mask_2 = os.path.join(mask_dir,mask_base_2)
            else:
                catalog = os.path.join(catalog_dir,catalog_base%pix)
                mask_1 = os.path.join(mask_dir,mask_base_1%pix)
                mask_2 = os.path.join(mask_dir,mask_base_2%pix)
            data[ii]['pix'] = pix if pix is not None else -1
            data[ii]['catalog'] = catalog
            data[ii]['mask_1']  = mask_1
            data[ii]['mask_2']  = mask_2
         
            mask[ii]['catalog'] = not os.path.exists(catalog)
            mask[ii]['mask_1']  = not os.path.exists(mask_1)
            mask[ii]['mask_2']  = not os.path.exists(mask_2)
        # 'pix' is masked if all files not present
        mask['pix'] = mask['catalog'] | mask['mask_1'] | mask['mask_2']
        #return numpy.ma.mrecords.MaskedArray(data, mask, fill_value=[-1,None,None,None])
        return numpy.ma.mrecords.MaskedArray(data, mask, fill_value=[-1,'','',''])

    getCatalogFiles = getFilenames

############################################################



#!/usr/bin/env python
"""
@author: Alex Drlica-Wagner <kadrlica@fnal.gov>
"""

import os
import copy
import subprocess
from collections import OrderedDict as odict

import healpy
import pyfits
import numpy as np
import numpy
import numpy.lib.recfunctions as recfuncs
import scipy.ndimage as ndimage

import ugali.candidate.associate
import ugali.utils.skymap

from ugali.utils.shell import mkdir, which
from ugali.utils.logger import logger
from ugali.utils.binning import reverseHistogram
from ugali.utils.projector import Projector, gal2cel, dec2hms, dec2dms, mod2dist
from ugali.utils.healpix import pix2ang, ang2pix

class CandidateSearch(object):
    """
    Object used to search for candidate objects in TS maps.
    """
    def __init__(self, config, mergefile=None, roifile=None):
        self.config = config
        self._config()
        if mergefile is not None: self.mergefile = mergefile
        if roifile is not None: self.roifile = roifile
        self._load()

    def _config(self):
        self.nside = self.config['coords']['nside_pixel']
        self.threshold = self.config['search']['obj_threshold'] # = 10
        self.xsize = self.config['search']['xsize'] # = 10000
        self.minpix = self.config['search']['minpix'] # = 1

        self.mergefile  = self.config.mergefile
        self.roifile    = self.config.roifile
        self.labelfile  = self.config.labelfile
        self.objectfile = self.config.objectfile
        self.assocfile  = self.config.assocfile
        self.candfile   = self.config.candfile

        mkdir(self.config['output']['searchdir'])
              
    def _load(self):
        self.loadLikelihood()
        self.loadROI()

    def loadLikelihood(self,filename=None):
        if filename is None: filename = self.mergefile
        f = pyfits.open(self.mergefile)
        self.pixels = f[1].data['PIX']
        self.values = 2*f[1].data['LOG_LIKELIHOOD']
        self.distances = f[2].data['DISTANCE_MODULUS']
        self.richness = f[1].data['RICHNESS']

    def loadROI(self,filename=None):
        if filename is None: filename = self.roifile
        self.ninterior = ugali.utils.skymap.readSparseHealpixMap(filename,'NINSIDE')
        self.nannulus = ugali.utils.skymap.readSparseHealpixMap(filename,'NANNULUS')
        self.stellar = ugali.utils.skymap.readSparseHealpixMap(filename,'STELLAR')

    def createLabels2D(self):
        """ 2D labeling at zmax """
        logger.debug("  Creating 2D labels...")
        self.zmax = np.argmax(self.values,axis=1)
        self.vmax = self.values[np.arange(len(self.pixels),dtype=int),self.zmax]

        kwargs=dict(pixels=self.pixels,values=self.vmax,nside=self.nside,
                    threshold=self.threshold,xsize=self.xsize)
        labels,nlabels = CandidateSearch.labelHealpix(**kwargs)
        self.nlabels = nlabels
        self.labels = np.repeat(labels,len(self.distances)).reshape(len(labels),len(self.distances))
        return self.labels, self.nlabels

    def createLabels3D(self):
        logger.debug("  Creating 3D labels...")
        kwargs=dict(pixels=self.pixels,values=self.values,nside=self.nside,
                    threshold=self.threshold,xsize=self.xsize)
        self.labels,self.nlabels = CandidateSearch.labelHealpix(**kwargs)
        return self.labels, self.nlabels

    def writeLabels(self,filename=None):
        if filename is None: filename = self.labelfile
        # Converting to float is a waste of memory...
        # This should be much more robustly done in writeSparseHealpixMap
        data_dict = {'LABEL':self.labels.astype(float)}
        logger.info("Writing %s..."%filename)
        ugali.utils.skymap.writeSparseHealpixMap(self.pixels,data_dict,self.nside,filename,
                                                 distance_modulus_array=self.distances)

    def loadLabels(self,filename=None):
        if filename is None: filename = self.labelfile
        f = pyfits.open(filename)
        if not (self.pixels == f[1].data['PIX']).all(): 
            raise Exception("...")
        if not (self.distances == f[2].data['DISTANCE_MODULUS']).all():
            raise Exception("...")            

        self.labels = f[1].data['LABEL'].astype(int)
        self.nlabels = self.labels.max()
        if self.nlabels != (len(np.unique(self.labels)) - 1):
            raise Exception("Incorrect number of labels found.")
        return self.labels, self.nlabels
        
    def createObjects(self):
        logger.debug("  Creating objects...")
        hist,edges,rev = reverseHistogram(self.labels,bins=numpy.arange(self.nlabels+2))
        self.rev = rev
        # Make some cut on the minimum size of a labelled object
        good, = numpy.where( (hist >= self.minpix) )
        # Get rid of zero label (below threshold)
        self.good, = numpy.nonzero(good)

        kwargs=dict(pixels=self.pixels,values=self.values,nside=self.nside, 
                    zvalues=self.distances, rev=self.rev, good=self.good)
        objects = self.findObjects(**kwargs)
        self.objects = self.finalizeObjects(objects)

        return self.objects

    def writeObjects(self,filename=None):
        if filename is None: filename = self.objectfile
        hdu = pyfits.new_table(self.objects.view(np.recarray))
        logger.info("Writing %s..."%filename)
        hdu.writeto(filename,clobber=True)

    @staticmethod
    def labelHealpix(pixels, values, nside, threshold=0, xsize=1000):
        """
        Label contiguous regions of a (sparse) HEALPix map. Works by mapping 
        HEALPix array to a Mollweide projection and applying scipy.ndimage.label
     
        Assumes non-nested HEALPix map.
        
        Parameters:
        pixels    : Pixel values associated to (sparse) HEALPix array
        values    : (Sparse) HEALPix array of data values
        nside     : HEALPix dimensionality
        threshold : Threshold value for object detection
        xsize     : Size of Mollweide projection
        
        Returns:
        labels, nlabels
        """
        proj = healpy.projector.MollweideProj(xsize=xsize)
        vec = healpy.pix2vec(nside,pixels)
        xy = proj.vec2xy(vec)
        ij = proj.xy2ij(xy)
        xx,yy = proj.ij2xy()
     
        # Convert to Mollweide
        searchims = []
        if values.ndim < 2: iterate = [values]
        else:               iterate = values.T
        for i,value in enumerate(iterate):
            logger.debug("Labeling slice %i...")
            searchim = numpy.zeros(xx.shape,dtype=bool)
            select = (value > threshold)
            yidx = ij[0][select]; xidx = ij[1][select]
            searchim[yidx,xidx] |= True
            searchims.append( searchim )
        searchims = numpy.array(searchims)

        # Full binary structure
        s = ndimage.generate_binary_structure(searchims.ndim,searchims.ndim)
     
        ### # Dilate in the z-direction
        logger.info("  Dilating image...")
        searchims = ndimage.binary_dilation(searchims,s,1)
        
        # Do the labeling
        logger.info("  Labeling image...")
        labels,nlabels = ndimage.label(searchims,structure=s)

        # Convert back to healpix
        pix_labels = labels[:,ij[0],ij[1]].T
        pix_labels = pix_labels.reshape(values.shape)
        pix_labels *= (values > threshold) # re-trim

        return pix_labels, nlabels

    @staticmethod
    def findObjects(pixels, values, nside, zvalues, rev, good):
        """
        Characterize labelled candidates in a multi-dimensional HEALPix map.
     
        Parameters:
        values    : (Sparse) HEALPix array of data values
        nside     : HEALPix dimensionality
        pixels    : Pixel values associated to (sparse) HEALPix array
        zvalues   : Values of the z-dimension (usually distance modulus)
        rev       : Reverse indices for pixels in each "island"
        good      : Array containg labels for each "island"
     
        Returns:
        objs      : numpy.recarray of object characteristics
        """
     
        ngood = len(good)
        objs = numpy.recarray((ngood,),
                           dtype=[('LABEL','i4'),
                                  ('NPIX','i4'),
                                  ('VAL_MAX','f4'),
                                  ('IDX_MAX','i4'),
                                  ('ZIDX_MAX','i4'),
                                  ('PIX_MAX','i4'),
                                  ('X_MAX','f4'),
                                  ('Y_MAX','f4'),
                                  ('Z_MAX','f4'),
                                  ('X_CENT','f4'),
                                  ('Y_CENT','f4'),
                                  ('Z_CENT','f4'),
                                  ('X_BARY','f4'),
                                  ('Y_BARY','f4'),
                                  ('Z_BARY','f4'),
                                  ('CUT','i2'),])
        objs['CUT'][:] = 0
     
        shape = values.shape
        ncol = shape[1]
        for i in range(0,ngood):
            logger.debug("i=%i",i)
            # This code could use some cleanup...
            indices=rev[rev[good[i]]:rev[good[i]+1]]
            npix = len(indices)
            idx = indices // ncol # This is the spatial index
            zidx = indices % ncol  # This is the distance index
     
            pix = pixels[idx] # This is the healpix pixel
            xval,yval = pix2ang(nside, pix)
            zval = zvalues[zidx]
            
            objs[i]['LABEL'] = good[i]
            objs[i]['NPIX'] = npix
            logger.debug("LABEL=%i"%objs[i]['LABEL'])
            logger.debug("NPIX=%i"%objs[i]['NPIX'])
     
            island = values[idx,zidx]
            idxmax = island.argmax()
            xval_max,yval_max,zval_max = xval[idxmax],yval[idxmax],zval[idxmax]
     
            objs[i]['VAL_MAX'] = island[idxmax]
            objs[i]['IDX_MAX']  = idx[idxmax]
            objs[i]['ZIDX_MAX']  = zidx[idxmax]
            objs[i]['PIX_MAX']   = pix[idxmax]
            objs[i]['X_MAX']  = xval_max
            objs[i]['Y_MAX']  = yval_max
            objs[i]['Z_MAX']  = zval_max

            proj = Projector(xval_max,yval_max)
            xpix,ypix = proj.sphereToImage(xval,yval)

            # Projected centroid
            x_cent,y_cent,zval_cent = numpy.average([xpix,ypix,zval],axis=1)
            xval_cent, yval_cent = proj.imageToSphere(x_cent,y_cent)
            objs[i]['X_CENT'] = xval_cent
            objs[i]['Y_CENT'] = yval_cent
            objs[i]['Z_CENT'] = zval_cent

            # Projected barycenter
            weights=[island,island,island]
            x_bary,y_bary,zval_bary = numpy.average([xpix,ypix,zval],weights=weights,axis=1)
            xval_bary,yval_bary = proj.imageToSphere(x_bary, y_bary)
            objs[i]['X_BARY'] = xval_bary
            objs[i]['Y_BARY'] = yval_bary
            objs[i]['Z_BARY'] = zval_bary
     
        return objs

    def finalizeObjects(self, objects):
        objs = numpy.recarray(len(objects),
                              dtype=[('NAME','S24'),
                                     ('TS','f4'),
                                     ('GLON','f4'),
                                     ('GLAT','f4'),
                                     ('RA','f4'),
                                     ('DEC','f4'),
                                     ('MODULUS','f4'),
                                     ('DISTANCE','f4'),
                                     ('RICHNESS','f4'),
                                     ('MASS','f4'),
                                     ('NANNULUS','i4'),
                                     ('NINTERIOR','i4'),
                                     ])
        
        objs['TS'] = self.values[objects['IDX_MAX'],objects['ZIDX_MAX']]

        glon,glat = objects['X_MAX'],objects['Y_MAX']
        objs['GLON'],objs['GLAT'] = glon,glat
        
        ra,dec    = gal2cel(glon,glat)
        objs['RA'],objs['DEC'] = ra,dec

        modulus = objects['Z_MAX']
        objs['MODULUS'] = modulus
        objs['DISTANCE'] = mod2dist(modulus)

        #ninterior = ugali.utils.skymap.readSparseHealpixMap(self.roifile,'NINSIDE')
        #nannulus = ugali.utils.skymap.readSparseHealpixMap(self.roifile,'NANNULUS')
        #stellar = ugali.utils.skymap.readSparseHealpixMap(self.roifile,'STELLAR')

        nside = healpy.npix2nside(len(self.nannulus))
        pix = ang2pix(nside,glon,glat)

        richness = self.richness[objects['IDX_MAX'],objects['ZIDX_MAX']]
        objs['RICHNESS'] = richness
        objs['MASS'] = richness * self.stellar[pix]

        objs['NANNULUS']  = self.nannulus[pix].astype(int)
        objs['NINTERIOR'] = self.ninterior[pix].astype(int)

        # Default name formatting
        # http://cdsarc.u-strasbg.fr/ftp/pub/iau/
        # http://cds.u-strasbg.fr/vizier/Dic/iau-spec.htx
        fmt = "J%(hour)02i%(hmin)04.1f%(deg)+03i%(dmin)02i"
        for obj,_ra,_dec in zip(objs,ra,dec):
            hms = dec2hms(_ra); dms = dec2dms(_dec)
            params = dict(hour=hms[0],hmin=hms[1]+hms[2]/60.,
                          deg=dms[0],dmin=dms[1]+dms[2]/60.)
            obj['NAME'] = fmt%params

        out = recfuncs.merge_arrays([objs,objects],usemask=False,asrecarray=True,flatten=True)
        # This is safer than viewing as FITS_rec
        return pyfits.new_table(out).data

    def loadLabels(self,filename=None):
        if filename is None: filename = self.labelfile
        f = pyfits.open(filename)
        if not (self.pixels == f[1].data['PIX']).all(): 
            raise Exception("...")
        if not (self.distances == f[2].data['DISTANCE_MODULUS']).all():
            raise Exception("...")            

        self.labels = f[1].data['LABEL'].astype(int)
        self.nlabels = self.labels.max()
        if self.nlabels != (len(np.unique(self.labels)) - 1):
            raise Exception("Incorrect number of labels found.")
        return self.labels, self.nlabels

    def loadObjects(self,filename=None):
        if filename is None: filename = self.objectfile
        f = pyfits.open(filename)
        self.objects = f[1].data

    def loadAssociations(self,filename=None):
        if filename is None: filename = self.assocfile
        f = pyfits.open(filename)
        self.assocs = f[1].data

    def createAssociations(self):
        objects = self.objects

        tol = self.config['search']['proximity']
        columns = []
     
        names = np.empty(len(objects),dtype=object)
        names.fill('')
        for i,refs in enumerate(self.config['search']['catalogs']):
            i += 1
            catalog = ugali.candidate.associate.SourceCatalog()
            for ref in refs:
                catalog += ugali.candidate.associate.catalogFactory(ref)
     
            # String length (should be greater than longest name)
            length = len(max(catalog['name'],key=len)) + 1
            dtype = 'S%i'%length; fitstype='%iA'%length
     
            assoc = np.empty(len(objects),dtype=dtype)
            assoc.fill('')
            idx1,idx2,dist = catalog.match(objects['GLON'],objects['GLAT'],tol=tol)
            assoc[idx1] = catalog['name'][idx2].astype(dtype)
            columns.append(pyfits.Column(name='ASSOC%i'%i,format=fitstype,array=assoc))
            columns.append(pyfits.Column(name='ANGSEP%i'%i,format='E',array=dist))

            if length > objects['NAME'].itemsize:
                logger.warning("Association name may not fit.")
            names = np.where(names=='',assoc,names)
        names = names.astype(objects['NAME'].dtype)
        objects['NAME'][:] = np.where(names=='',objects['NAME'],names)

        self.assocs = pyfits.new_table(objects.columns+pyfits.ColDefs(columns)).data
        self.assocs = self.assocs[self.assocs['NAME'].argsort()]

    def writeAssociations(self,filename=None):
        if filename is None: filename = self.assocfile
        hdu = pyfits.new_table(self.assocs.view(np.recarray))
        logger.info("Writing %s..."%filename)
        hdu.writeto(filename,clobber=True)

    def writeCandidates(self,filename=None):
        if filename is None: filename = self.candfile

        threshold = self.config['search']['cand_threshold']
        select  = (self.assocs['CUT']==0)
        select &= (self.assocs['TS']>threshold)
        #select &= (self.assocs['ASSOC2']=='')

        self.candidates = self.assocs[select]
        # ADW: View as a recarray or selection doesn't work.
        # Why? I don't know, and I'm slightly terrified...
        hdu = pyfits.new_table(self.candidates.view(np.recarray))
        logger.info("Writing %s..."%filename)
        hdu.writeto(filename,clobber=True)

        # Dump to txt file 
        if which('fdump'):
            txtfile = filename.replace('.fits','.txt')
            columns = ['NAME','TS','GLON','GLAT','DISTANCE','MASS']
            cmd = 'fdump %(infile)s %(outfile)s columns="%(columns)s" rows="-" prhead="no" showcol="yes" clobber="yes" pagewidth="256" fldsep=" " showrow="no"'%(dict(infile=filename,outfile=txtfile,columns=','.join(columns)))
            print cmd
            subprocess.call(cmd,shell=True)

if __name__ == "__main__":
    from optparse import OptionParser
    usage = "Usage: %prog  [options] input"
    description = "python script"
    parser = OptionParser(usage=usage,description=description)
    (opts, args) = parser.parse_args()



"""
Functions to create and use a look-up table for the signal color probability distribution function.
"""

import time
import numpy
import scipy.signal
import pyfits

import ugali.utils.config
import ugali.utils.binning
import ugali.analysis.isochrone

from ugali.utils.logger import logger

############################################################

def writeColorLUT2(config,
                  outfile=None, isochrone=None, distance_modulus_array=None,
                  delta_mag=None, mag_err_array=None,
                  mass_steps=10000, plot=False):
    """
    Precompute a 4-dimensional signal color probability look-up table to speed up the likelihood evaluation.
    Inputs are a Config object (or file name), an Isochrone object,
    an array of distance moduli at which to evaluate the signal color probability,
    and an array of magnitude uncertainties which set the bin edges of those dimensions (zero implicity included).
    Finally there is an outfile name.
    """
    if plot: import ugali.utils.plotting

    if type(config) == str:
        config = ugali.utils.config.Config(config)
    if outfile is None:
        outfile = config.params['color_lut']['filename']
    if isochrone is None:
        isochrones = []
        for ii, name in enumerate(config.params['isochrone']['infiles']):
            isochrones.append(ugali.analysis.isochrone.Isochrone(config, name))
        isochrone = ugali.analysis.isochrone.CompositeIsochrone(isochrones, config.params['isochrone']['weights'])
    if distance_modulus_array is None:
        distance_modulus_array = config.params['color_lut']['distance_modulus_array']
    if delta_mag is None:
        delta_mag = config.params['color_lut']['delta_mag']
    if mag_err_array is None:
        mag_err_array = config.params['color_lut']['mag_err_array']

    mag_buffer = 0.5 # Safety buffer in magnitudes around the color-magnitude space defined by the ROI
    epsilon = 1.e-10
    if config.params['catalog']['band_1_detection']:
        bins_mag_1 = numpy.arange(config.params['mag']['min'] - mag_buffer,
                                  config.params['mag']['max'] + mag_buffer + epsilon,
                                  delta_mag)
        bins_mag_2 = numpy.arange(config.params['mag']['min'] - config.params['color']['max'] - mag_buffer,
                                  config.params['mag']['max'] - config.params['color']['min'] + mag_buffer + epsilon,
                                  delta_mag)
    else:
        bins_mag_1 = numpy.arange(config.params['mag']['min'] + config.params['color']['min'] - mag_buffer,
                                  config.params['mag']['max'] + config.params['color']['max'] + mag_buffer + epsilon,
                                  delta_mag)
        bins_mag_2 = numpy.arange(config.params['mag']['min'] - mag_buffer,
                                  config.params['mag']['max'] + mag_buffer + epsilon,
                                  delta_mag)

    # Output binning configuration
    #print config.params['catalog']['band_1_detection']
    #print config.params['mag']['min'], config.params['mag']['max']
    #print config.params['color']['min'], config.params['color']['max']

    #print bins_mag_1[0], bins_mag_1[-1], len(bins_mag_1)
    #print bins_mag_2[0], bins_mag_2[-1], len(bins_mag_2)

    isochrone_mass_init, isochrone_mass_pdf, isochrone_mass_act, isochrone_mag_1, isochrone_mag_2 = isochrone.sample(mass_steps=mass_steps)

    hdul = pyfits.HDUList()

    for index_distance_modulus, distance_modulus in enumerate(distance_modulus_array):

        logger.debug('(%i/%i)'%(index_distance_modulus, len(distance_modulus_array)))

        columns_array = []
        
        time_start = time.time()

        histo_isochrone_pdf = numpy.histogram2d(distance_modulus + isochrone_mag_1,
                                                distance_modulus + isochrone_mag_2,
                                                bins=[bins_mag_1, bins_mag_2],
                                                weights=isochrone_mass_pdf)[0]
        
        if plot:
            # Checked that axis are plotted correctly
            ugali.utils.plotting.twoDimensionalHistogram('Isochrone', 'mag_1', 'mag_2',
                                                         numpy.log10(histo_isochrone_pdf + epsilon).transpose(),
                                                         bins_mag_1, bins_mag_2,
                                                         lim_x=None, lim_y=None,
                                                         vmin=None, vmax=None)

            
        
        for index_mag_err_1, mag_err_1 in enumerate(mag_err_array):
            for index_mag_err_2, mag_err_2 in enumerate(mag_err_array):
                logger.debug('  Distance modulus = %.2f mag_err_1 = %.2f mag_err_2 = %.2f'%(distance_modulus, mag_err_1, mag_err_2))

                mag_1_sigma_step = delta_mag / mag_err_1
                n = int(numpy.ceil(4. / mag_1_sigma_step))
                mag_1_sigma = numpy.arange(-1. * (n + 0.5) * mag_1_sigma_step,
                                           ((n + 0.5) * mag_1_sigma_step) + epsilon,
                                           mag_1_sigma_step)
                mag_1_pdf_array = scipy.stats.norm.cdf(mag_1_sigma[1:]) - scipy.stats.norm.cdf(mag_1_sigma[0:-1])

                mag_2_sigma_step = delta_mag / mag_err_2
                n = int(numpy.ceil(4. / mag_2_sigma_step))
                mag_2_sigma = numpy.arange(-1. * (n + 0.5) * mag_2_sigma_step,
                                           ((n + 0.5) * mag_2_sigma_step) + epsilon,
                                           mag_2_sigma_step)
                mag_2_pdf_array = scipy.stats.norm.cdf(mag_2_sigma[1:]) - scipy.stats.norm.cdf(mag_2_sigma[0:-1])

                mag_1_pdf, mag_2_pdf = numpy.meshgrid(mag_2_pdf_array, mag_1_pdf_array)
                
                pdf = mag_1_pdf * mag_2_pdf

                histo_isochrone_pdf_convolve = scipy.signal.convolve2d(histo_isochrone_pdf, pdf, mode='same')

                if plot:
                    # Checked that axis are plotted correctly
                    ugali.utils.plotting.twoDimensionalHistogram('Convolved Isochrone', 'mag_1', 'mag_2',
                                                                 numpy.log10(histo_isochrone_pdf_convolve + epsilon).transpose(),
                                                                 bins_mag_1, bins_mag_2,
                                                                 lim_x=None, lim_y=None,
                                                                 vmin=None, vmax=None)

                columns_array.append(pyfits.Column(name = '%i%i'%(index_mag_err_1, index_mag_err_2),
                                                   format = '%iE'%(histo_isochrone_pdf_convolve.shape[1]),
                                                   array = histo_isochrone_pdf_convolve))

        hdu = pyfits.new_table(columns_array)
        hdu.header.update('DIST_MOD', distance_modulus)
        hdu.name = '%.2f'%(distance_modulus)
        hdul.append(hdu)

        time_end = time.time()
        logger.debug('%.2f s'%(time_end - time_start))

    # Store distance modulus info
    columns_array = [pyfits.Column(name = 'DISTANCE_MODULUS',
                                   format = 'E',
                                   array = distance_modulus_array)]
    hdu = pyfits.new_table(columns_array)
    hdu.name = 'DISTANCE_MODULUS'
    hdul.append(hdu)

    # Store magnitude error info
    columns_array = [pyfits.Column(name = 'BINS_MAG_ERR',
                                   format = 'E',
                                   array = numpy.insert(mag_err_array, 0, 0.))]
    hdu = pyfits.new_table(columns_array)
    hdu.name = 'BINS_MAG_ERR'
    hdul.append(hdu)

    # Store magnitude 1 info
    columns_array = [pyfits.Column(name = 'BINS_MAG_1',
                                   format = 'E',
                                   array = bins_mag_1)]
    hdu = pyfits.new_table(columns_array)
    hdu.name = 'BINS_MAG_1'
    hdul.append(hdu)

    # Store magnitude 2 info
    columns_array = [pyfits.Column(name = 'BINS_MAG_2',
                                   format = 'E',
                                   array = bins_mag_2)]
    hdu = pyfits.new_table(columns_array)
    hdu.name = 'BINS_MAG_2'
    hdul.append(hdu)

    logger.info('Writing look-up table to %s'%(outfile))
    hdul.writeto(outfile, clobber = True)

############################################################

def writeColorLUT(config,
                  outfile=None, isochrone=None, distance_modulus_array=None,
                  delta_mag=None, mag_err_array=None,
                  mass_steps=1000000, plot=False):
    """
    Precompute a 4-dimensional signal color probability look-up table to speed up the likelihood evaluation.
    Inputs are a Config object (or file name), an Isochrone object,
    an array of distance moduli at which to evaluate the signal color probability,
    and an array of magnitude uncertainties which set the bin edges of those dimensions (zero implicity included).
    Finally there is an outfile name.
    """
    if plot: import ugali.utils.plotting
    if type(config) == str:
        config = ugali.utils.config.Config(config)
    if outfile is None:
        outfile = config.params['color_lut']['filename']
    if isochrone is None:
        isochrones = []
        for ii, name in enumerate(config.params['isochrone']['infiles']):
            isochrones.append(ugali.analysis.isochrone.Isochrone(config, name))
        isochrone = ugali.analysis.isochrone.CompositeIsochrone(isochrones, config.params['isochrone']['weights'])
    if distance_modulus_array is None:
        distance_modulus_array = config.params['color_lut']['distance_modulus_array']
    if delta_mag is None:
        delta_mag = config.params['color_lut']['delta_mag']
    if mag_err_array is None:
        mag_err_array = config.params['color_lut']['mag_err_array']

    mag_buffer = 0.5 # Safety buffer in magnitudes around the color-magnitude space defined by the ROI
    epsilon = 1.e-10
    if config.params['catalog']['band_1_detection']:
        bins_mag_1 = numpy.arange(config.params['mag']['min'] - mag_buffer,
                                  config.params['mag']['max'] + mag_buffer + epsilon,
                                  delta_mag)
        bins_mag_2 = numpy.arange(config.params['mag']['min'] - config.params['color']['max'] - mag_buffer,
                                  config.params['mag']['max'] - config.params['color']['min'] + mag_buffer + epsilon,
                                  delta_mag)
    else:
        bins_mag_1 = numpy.arange(config.params['mag']['min'] + config.params['color']['min'] - mag_buffer,
                                  config.params['mag']['max'] + config.params['color']['max'] + mag_buffer + epsilon,
                                  delta_mag)
        bins_mag_2 = numpy.arange(config.params['mag']['min'] - mag_buffer,
                                  config.params['mag']['max'] + mag_buffer + epsilon,
                                  delta_mag)

    # Output binning configuration
    #print config.params['catalog']['band_1_detection']
    #print config.params['mag']['min'], config.params['mag']['max']
    #print config.params['color']['min'], config.params['color']['max']

    #print bins_mag_1[0], bins_mag_1[-1], len(bins_mag_1)
    #print bins_mag_2[0], bins_mag_2[-1], len(bins_mag_2)

    isochrone_mass_init, isochrone_mass_pdf, isochrone_mass_act, isochrone_mag_1, isochrone_mag_2 = isochrone.sample(mass_steps=mass_steps)

    # make randoms
    randoms_1 = numpy.random.normal(0., 1., len(isochrone_mass_pdf))
    randoms_2 = numpy.random.normal(0., 1., len(isochrone_mass_pdf))

    hdul = pyfits.HDUList()

    for index_distance_modulus, distance_modulus in enumerate(distance_modulus_array):

        logger.debug('(%i/%i)'%(index_distance_modulus, len(distance_modulus_array)))

        columns_array = []
        
        time_start = time.time()
        
        for index_mag_err_1, mag_err_1 in enumerate(mag_err_array):
            for index_mag_err_2, mag_err_2 in enumerate(mag_err_array):
                logger.debug('  (%i/%i) Distance modulus = %.2f mag_err_1 = %.3f mag_err_2 = %.3f'%(index_mag_err_1 * len(mag_err_array) + index_mag_err_2,
                                                                                             len(mag_err_array)**2,
                                                                                             distance_modulus,
                                                                                             mag_err_1,
                                                                                             mag_err_2))
                
                # Add randoms
                histo_isochrone_pdf = numpy.histogram2d(distance_modulus + isochrone_mag_1 + randoms_1 * mag_err_1,
                                                        distance_modulus + isochrone_mag_2 + randoms_2 * mag_err_2,
                                                        bins=[bins_mag_1, bins_mag_2],
                                                        weights=isochrone_mass_pdf)[0]
                
                if plot:
                    # Checked that axis are plotted correctly
                    ugali.utils.plotting.twoDimensionalHistogram('Convolved Isochrone', 'mag_1', 'mag_2',
                                                                 numpy.log10(histo_isochrone_pdf + epsilon).transpose(),
                                                                 bins_mag_1, bins_mag_2,
                                                                 lim_x=None, lim_y=None,
                                                                 vmin=None, vmax=None)
                    raw_input('WAIT')

                columns_array.append(pyfits.Column(name = '%i%i'%(index_mag_err_1, index_mag_err_2),
                                                   format = '%iE'%(histo_isochrone_pdf.shape[1]),
                                                   array = histo_isochrone_pdf))

        hdu = pyfits.new_table(columns_array)
        hdu.header.update('DIST_MOD', distance_modulus)
        hdu.name = '%.2f'%(distance_modulus)
        hdul.append(hdu)

        time_end = time.time()
        logger.debug('%.2f s'%(time_end - time_start))

    # Store distance modulus info
    columns_array = [pyfits.Column(name = 'DISTANCE_MODULUS',
                                   format = 'E',
                                   array = distance_modulus_array)]
    hdu = pyfits.new_table(columns_array)
    hdu.name = 'DISTANCE_MODULUS'
    hdul.append(hdu)

    # Store magnitude error info
    columns_array = [pyfits.Column(name = 'BINS_MAG_ERR',
                                   format = 'E',
                                   array = numpy.insert(mag_err_array, 0, 0.))]
    hdu = pyfits.new_table(columns_array)
    hdu.name = 'BINS_MAG_ERR'
    hdul.append(hdu)

    # Store magnitude 1 info
    columns_array = [pyfits.Column(name = 'BINS_MAG_1',
                                   format = 'E',
                                   array = bins_mag_1)]
    hdu = pyfits.new_table(columns_array)
    hdu.name = 'BINS_MAG_1'
    hdul.append(hdu)

    # Store magnitude 2 info
    columns_array = [pyfits.Column(name = 'BINS_MAG_2',
                                   format = 'E',
                                   array = bins_mag_2)]
    hdu = pyfits.new_table(columns_array)
    hdu.name = 'BINS_MAG_2'
    hdul.append(hdu)

    logger.info('Writing look-up table to %s'%(outfile))
    hdul.writeto(outfile, clobber = True)

############################################################

def mergeColorLUT(infiles):
    """
    Tool to merge color look-up tables.
    """
    pass

############################################################

def readColorLUT(infile, distance_modulus, mag_1, mag_2, mag_err_1, mag_err_2):
    """
    Take in a color look-up table and return the signal color evaluated for each object.
    Consider making the argument a Catalog object rather than magnitudes and uncertainties.
    """

    reader = pyfits.open(infile)

    distance_modulus_array = reader['DISTANCE_MODULUS'].data.field('DISTANCE_MODULUS')
    if not numpy.any(numpy.fabs(distance_modulus_array - distance_modulus) < 1.e-3):
        logger.warning("Distance modulus %.2f not available in file %s"%(distance_modulus, infile))
        logger.warning('         available distance moduli:'+str(distance_modulus_array))
        return False

    distance_modulus_key = '%.2f'%(distance_modulus_array[numpy.argmin(numpy.fabs(distance_modulus_array - distance_modulus))])

    bins_mag_err = reader['BINS_MAG_ERR'].data.field('BINS_MAG_ERR') 
    bins_mag_1 = reader['BINS_MAG_1'].data.field('BINS_MAG_1') 
    bins_mag_2 = reader['BINS_MAG_2'].data.field('BINS_MAG_2') 
    # Note that magnitude uncertainty is always assigned by rounding up, is this the right thing to do?
    index_mag_err_1 = numpy.clip(numpy.digitize(mag_err_1, bins_mag_err) - 1,
                                 0, len(bins_mag_err) - 2)
    index_mag_err_2 = numpy.clip(numpy.digitize(mag_err_2, bins_mag_err) - 1,
                                 0, len(bins_mag_err) - 2)

    u_color = numpy.zeros(len(mag_1))

    for index_mag_err_1_select in range(0, len(bins_mag_err) - 1):
        for index_mag_err_2_select in range(0, len(bins_mag_err) - 1):
            cut = numpy.logical_and(index_mag_err_1 == index_mag_err_1_select,
                                    index_mag_err_2 == index_mag_err_2_select)
            if numpy.sum(cut) < 1:
                continue
            histo = reader[distance_modulus_key].data.field('%i%i'%(index_mag_err_1_select, index_mag_err_2_select))
            u_color[cut] = ugali.utils.binning.take2D(histo,
                                                      mag_2[cut], mag_1[cut],
                                                      bins_mag_2, bins_mag_1)
    
    reader.close()
    return u_color

############################################################



#!/usr/bin/env python

"""
Class to farm out analysis tasks.

@author: Keith Bechtol      <bechtol@kicp.uchicago.edu>
@author: Alex Drlica-Wagner <kadrlica@fnal.gov>
"""

import os
from os.path import join, exists
import sys
import subprocess
import time
import glob

import numpy
import numpy as np
import healpy

import ugali.utils.config
import ugali.utils.skymap
import ugali.utils.batch

from ugali.utils.projector import gal2cel,cel2gal
from ugali.utils.healpix import subpixel,superpixel,query_disc
from ugali.utils.healpix import pix2ang,ang2vec
from ugali.utils.logger import logger
from ugali.utils.shell import mkdir

class Farm:

    def __init__(self, configfile):
        self.configfile = configfile
        self.config = ugali.utils.config.Config(configfile)
        self._setup()

    def _setup(self):
        self.nside_catalog    = self.config['coords']['nside_catalog']
        self.nside_likelihood = self.config['coords']['nside_likelihood']
        self.nside_pixel      = self.config['coords']['nside_pixel']

        self.filenames = self.config.getFilenames()
        self.skip = "Outfile already exists. Skipping..."

        # Might consider storing only the good filenames
        # self.filenames = self.filenames.compress(~self.filenames.mask['pix'])
        self.catalog_pixels = self.filenames['pix'].compressed()

    def command(self, outfile, configfile, pix):
        """
        Generate the command for running the likelihood scan.
        """
        params = dict(script=self.config['scan']['script'],
                      config=configfile, outfile=outfile, 
                      nside=self.nside_likelihood, pix=pix)
        cmd = '%(script)s %(config)s %(outfile)s --hpx %(nside)i %(pix)i'%params
        return cmd

    # ADW: Should probably be in a utility
    def footprint(self, nside=None):
        """
        UNTESTED.
        Should return a boolean array representing the pixels in the footprint.
        """
        if nside is None:
            nside = self.nside_pixel
        elif nside < self.nside_catalog: 
            raise Exception('Requested nside=%i is less than catalog_nside'%nside)
        elif nside > self.nside_pixel:
            raise Exception('Requested nside=%i is greater than pixel_nside'%nside)
        pix = numpy.arange(healpy.nside2npix(nside), dtype=int)
        map = self.inFootprint(pix,nside)
        return map 

    # ADW: Should probably be in a utility
    def inFootprint(self, pixels, nside=None):
        """
        Open each valid filename for the set of pixels and determine the set 
        of subpixels with valid data.
        """
        if numpy.isscalar(pixels): pixels = numpy.array([pixels])
        if nside is None: nside = self.nside_likelihood

        inside = numpy.zeros( len(pixels), dtype='bool')
        if not self.nside_catalog:
            catalog_pix = [0]
        else:
            catalog_pix = superpixel(pixels,nside,self.nside_catalog)
            catalog_pix = numpy.intersect1d(catalog_pix,self.catalog_pixels)

        for filenames in self.filenames[catalog_pix]:
            #logger.debug("Loading %s"%filenames['mask_1'])
            subpix_1,val_1 = ugali.utils.skymap.readSparseHealpixMap(filenames['mask_1'],'MAGLIM',construct_map=False)
            #logger.debug("Loading %s"%filenames['mask_2'])
            subpix_2,val_2 = ugali.utils.skymap.readSparseHealpixMap(filenames['mask_2'],'MAGLIM',construct_map=False)
            subpix = numpy.intersect1d(subpix_1,subpix_2)
            superpix = numpy.unique(ugali.utils.skymap.superpixel(subpix,self.nside_pixel,nside))
            inside |= numpy.in1d(pixels, superpix)
            
        return inside
        
    def submit_all(self, coords=None, queue=None, debug=False):
        """
        Submit likelihood analyses on a set of coordinates. If
        coords is `None`, submit all coordinates in the footprint.

        Inputs:
        coords : Array of target locations in Galactic coordinates. 
        queue  : Overwrite submit queue.
        debug  : Don't run.
        """
        if coords is None:
            pixels = numpy.arange(healpy.nside2npix(self.nside_likelihood))
        else:
            coords = numpy.asarray(coords)
            if coords.ndim == 1:
                coords = numpy.array([coords])
            if coords.shape[1] == 2:
                glon,glat = coords.T
                radius    = numpy.zeros(len(glon))
            elif coords.shape[1] == 3:
                glon,glat,radius = coords.T
            else:
                raise Exception("Unrecognized coords shape:"+str(coords.shape))
            vec = ang2vec(glon,glat)
            pixels = numpy.zeros(0, dtype=int)
            for v,r in zip(vec,radius):
                pix = query_disc(self.nside_likelihood,v,r,inclusive=True,fact=32)
                pixels = numpy.hstack([pixels, pix])
            #pixels = numpy.unique(pixels)

        inside = ugali.utils.skymap.inFootprint(self.config,pixels)
        if inside.sum() != len(pixels):
            logger.warning("Ignoring pixels outside survey footprint:\n"+str(pixels[~inside]))
        if inside.sum() == 0:
            logger.warning("No pixels inside footprint.")
            return

        # Only write the configfile once
        outdir = mkdir(self.config['output']['likedir'])
        configfile = '%s/config_queue.py'%(outdir)
        self.config.write(configfile)

        pixels = pixels[inside]
        self.submit(pixels,queue=queue,debug=debug,configfile=configfile)

    def submit(self, pixels, queue=None, debug=False, configfile=None):
        """
        Submit the likelihood job for the given pixel(s).
        """
        queue = self.config['batch']['cluster'] if queue is None else queue
        local = (queue == 'local')

        # Need to develop some way to take command line arguments...
        self.batch = ugali.utils.batch.batchFactory(queue,**self.config['batch']['opts'])

        if numpy.isscalar(pixels): pixels = numpy.array([pixels])

        outdir = mkdir(self.config['output']['likedir'])
        logdir = mkdir(join(outdir,'log'))
        subdir = mkdir(join(outdir,'sub'))

        # Save the current configuation settings; avoid writing 
        # file multiple times if configfile passed as argument.
        if configfile is None:
            if local:
                configfile = self.configfile
            else:
                configfile = '%s/config_queue.py'%(outdir)
                self.config.write(configfile)
                
        lon,lat = pix2ang(self.nside_likelihood,pixels)
        commands = []
        chunk = self.config['batch']['chunk']
        istart = 0
        logger.info('=== Submit Likelihood ===')
        for ii,pix in enumerate(pixels):
            logger.info('  (%i/%i) pixel=%i nside=%i; (lon, lat) = (%.2f, %.2f)'%(ii+1,len(pixels),pix, self.nside_likelihood,lon[ii],lat[ii]))

            # Create outfile name
            outfile = self.config.likefile%(pix,self.config['coords']['coordsys'].lower())
            outbase = os.path.basename(outfile)
            jobname = self.config['batch']['jobname']

            # Submission command
            sub = not os.path.exists(outfile)
            cmd = self.command(outfile,configfile,pix)
            commands.append([ii,cmd,lon[ii],lat[ii],sub])
            
            if local or chunk == 0:
                # Not chunking
                command = cmd
                submit = sub
                logfile = join(logdir,os.path.splitext(outbase)[0]+'.log')
            elif (len(commands)%chunk==0) or (ii+1 == len(pixels)):
                # End of chunk, create submission script
                commands = np.array(commands,dtype=object)
                istart, iend = commands[0][0], commands[-1][0]
                subfile = join(subdir,'submit_%08i_%08i.sh'%(istart,iend))
                logfile = join(logdir,'submit_%08i_%08i.log'%(istart,iend))
                command = "sh %s"%subfile

                submit = np.any(commands[:,-1])
                if submit: self.write_script(subfile,commands)
            else:
                # Not end of chunk
                continue
            commands=[]

            # Actual job submission
            if not submit:
                logger.info(self.skip)
                continue
            else:
                while True:
                    njobs = self.batch.njobs()
                    if njobs < self.config['batch']['max_jobs']:
                        break
                    else:
                        logger.info('%i jobs already in queue, waiting...'%(njobs))
                        time.sleep(5*chunk)

                job = self.batch.submit(command,jobname,logfile)
                logger.info("  "+job)
                time.sleep(0.5)

    def write_script(self, filename, commands):
        info = 'echo "{0:=^60}";\n'
        hline = info.format("")
        newline = 'echo;\n'

        istart, iend = commands[0][0], commands[-1][0]
        script = open(filename,'w')
        script.write(hline)
        script.write(info.format('Submit Pixels %i to %i'%(istart,iend)))
        script.write(hline)
        script.write(newline)
        script.write('status=0;\n')
        for i,cmd,lon,lat,sub in commands: 
            script.write(info.format('Pixel %i: (%.2f, %.2f)'%(i,lon,lat)))
            if sub: script.write('%s; [ $? -ne 0 ] && status=1;\n'%cmd)
            else:   script.write('echo "%s";\n'%self.skip)
            script.write(hline)
            script.write(newline)
        script.write('exit $status;\n')
        script.close()

if __name__ == "__main__":
    import ugali.utils.parser
    description = "Script for dispatching the likelihood scan to the queue."
    parser = ugali.utils.parser.Parser(description=description)
    parser.add_config()
    parser.add_debug()
    parser.add_queue()
    parser.add_verbose()
    parser.add_coords(required=True,radius=True,targets=True)
    opts = parser.parse_args()

    farm = Farm(opts.config)
    x = farm.submit_all(coords=opts.coords,queue=opts.queue,debug=opts.debug)



#!/usr/bin/env python

import numpy
import numpy as np
import sys
import subprocess
import re
import os
import httplib
import StringIO

from ugali.utils.logger import logger
from ugali.utils.shell import mkdir

DATABASES = {
    'sdss':['dr10'],
    'des' :['sva1','sva1_gold','y1a1','y2n','y2q1'],
    }

def databaseFactory(survey,release):
    if survey == 'sdss':
        return SDSSDatabase(release = release)
    elif survey == 'des':
        if release == 'y1a1':
            return Y1A1Database()
        elif release == 'y2n':
            return Y2NDatabase()
        elif release == 'y2q1':
            return Y2Q1Database()
        else:
            msg = "Unrecognized release: %s"%release
            raise Exception(msg)
        return DESDatabase(release = release)
    else:
        logger.error("Unrecognized survey: %s"%survey)
        return None

class Database(object):
    def __init__(self):
        pass

    def load_pixels(self, pixfile=None):
        if pixfile is not None:
            self.pixels = np.loadtxt(pixfile,dtype=[('name',int),('ra_min',float),('ra_max',float),
                                                    ('dec_min',float),('dec_max',float)])
            # One-line input file
            if self.pixels.ndim == 0: self.pixels = np.array([self.pixels])
        else:
            self.pixels = self.create_pixels()

    @staticmethod
    def create_pixels(nra=18, ndec=10):
            ra_step = 360./nra #20
            ra_range = np.around(np.arange(0,360+ra_step,ra_step),0)
            # Decreasing dec...
            sin_dec_step = -(2.0/ndec) # -0.2
            dec_range = np.around(np.degrees(np.arcsin(np.arange(1,-1+sin_dec_step,sin_dec_step))),0)

            xx, yy = np.meshgrid( ra_range,dec_range)
            ra_min  = xx[1:,:-1].flatten(); ra_max = xx[1:,1:].flatten()
            dec_min = yy[1:,1:].flatten(); dec_max = yy[:-1,1:].flatten() # Decreasing...
            name = np.arange(len(ra_min),dtype=int)
            return  np.rec.fromarrays([name, ra_min, ra_max, dec_min, dec_max],
                                      dtype=[('name',int),('ra_min',float),('ra_max',float),
                                             ('dec_min',float),('dec_max',float)])


    def generate_query(self):
        """ Should be implemented by child class. """
        pass

    def download(self):
        """ Should be implemented by child class. """        
        pass

    def run(self):
        """ Should be implemented by child class. """
        pass

class SDSSDatabase(Database):
    """
    For downloading SDSS DR10 data set.
    """
    def __init__(self,release='DR10'):
        super(SDSSDatabase,self).__init__()
        self.release = release.lower()
        self.basename = "sdss_%s_photometry"%self.release

    def _setup_casjobs(self):
        # Function here to install casjobs.jar and CasJobs.config...
        # "wget http://skyserver.sdss3.org/CasJobs/download/casjobs.jar"
        # "wget http://skyserver.sdss3.org/CasJobs/download/CasJobs.config.x -o CasJobs.config"

        # For now, just check that they exist
        files = ['casjobs.jar','CasJobs.config']
        for f in files:
            if not os.path.exists(f):
                msg = "Can't find: %s"%f
                raise IOError(msg)

    def generate_query(self, ra_min,ra_max,dec_min,dec_max,filename,db):
        outfile = open(filename,"w")
        outfile.write('SELECT s.objID, s.ra AS "RA", s.dec as "DEC",\n')
        outfile.write('s.psfmag_g AS "MAG_PSF_G",\n')
        outfile.write('s.psfmagerr_g AS "MAGERR_PSF_G",\n')
        outfile.write('s.psfmag_g - s.extinction_g AS "MAG_PSF_SFD_G",\n')
        outfile.write('s.psfmag_r AS "MAG_PSF_R",\n')
        outfile.write('s.psfmagerr_r AS "MAGERR_PSF_R",\n')
        outfile.write('s.psfmag_r - s.extinction_r AS "MAG_PSF_SFD_R",\n')
        outfile.write('s.psfmag_i AS "MAG_PSF_I",\n')
        outfile.write('s.psfmagerr_i AS "MAGERR_PSF_I",\n')
        outfile.write('s.psfmag_i - s.extinction_i AS "MAG_PSF_SFD_I",\n')
        outfile.write('s.psfmag_z AS "MAG_PSF_Z",\n')
        outfile.write('s.psfmagerr_z AS "MAGERR_PSF_Z",\n')
        outfile.write('s.psfmag_z - s.extinction_z AS "MAG_PSF_SFD_Z"\n')
        outfile.write('INTO MyDB.%s\n' % (db))
        outfile.write('FROM %s.StarTag as s\n'%(self.release))
        outfile.write('WHERE s.ra > %.7f AND s.ra < %.7f\n' % (ra_min,ra_max))
        outfile.write('AND s.dec > %.7f AND s.dec < %.7f\n' % (dec_min,dec_max))
        outfile.write('AND s.clean = 1\n')
        outfile.close()

    def query(self,dbase,task,query):
        logger.info("Running query...")
        cmd = "java -jar casjobs.jar run -t %s -n %s -f %s" % (dbase,task,query)
        logger.info(cmd)
        ret = subprocess.check_output(cmd,shell=True,stderr=subprocess.STDOUT) 
        if 'ERROR:' in ret:
            raise subprocess.CalledProcessError(1,cmd,ret)
        return ret
        
    def extract(self, table):
        logger.info("Extracting...")
        cmd = "java -jar casjobs.jar extract -u -F -a FITS -b %s" % (table)
        logger.info(cmd)
        retval = subprocess.check_output(cmd,shell=True)

        url = None
        match = re.search("(http\:\/\/.*\.fit)",retval)
        if (match is not None) :
            url = match.group(0)
        else:
            logger.info("URL not found...here's the output")
            logger.info(retval)
        return url

    def wget(self,url,outfile=None):
        logger.info("Downloading %s\n" % (url))
        if outfile is not None: cmd = "wget -O %s %s" % (outfile,url)
        else:                   cmd = "wget %s" % (url)
        logger.info(cmd)
        return subprocess.check_output(cmd,shell=True)

    def drop(self, table):
        logger.info("Dropping...")
        cmd = "java -jar casjobs.jar execute -t MyDB -n \"drop query\" \"drop table %s\""%(table)
        logger.info(cmd)
        return subprocess.check_output(cmd,shell=True)

    def download(self, pixel, outdir=None, force=False):
        if outdir is None: outdir = './'
        else:              mkdir(outdir)
        sqldir = mkdir(os.path.join(outdir,'sql'))
        self._setup_casjobs()

        basename = self.basename + "_%04d"%pixel['name']
        sqlname = os.path.join(sqldir,basename+'.sql')
        dbname = basename+'_output'
        taskname = basename
        outfile = os.path.join(outdir,basename+".fits")
        if os.path.exists(outfile) and not force:
            logger.warning("Found %s; skipping..."%(outfile))
            return

        logger.info("\nDownloading pixel: %(name)i (ra=%(ra_min)g:%(ra_max)g,dec=%(dec_min)g:%(dec_max)g)"%(pixel))
        logger.info("Working on "+sqlname)
         
        self.generate_query(pixel['ra_min'],pixel['ra_max'],pixel['dec_min'],pixel['dec_max'],sqlname,dbname)

        try:
            self.query(self.release,taskname,sqlname)
        except subprocess.CalledProcessError, e:
            logger.error(e.output)
            self.drop(dbname)
            raise e
        
        try:
            url = self.extract(dbname)
        except subprocess.CalledProcessError, e:
            self.drop(dbname)
            raise e
            
        if (url is not None):
            self.wget(url,outfile)

        self.drop(dbname)

    def upload(self, array, fields=None, table="MyDB", configfile=None):
        """
        Upload an array to a personal database using SOAP POST protocol.
        http://skyserver.sdss3.org/casjobs/services/jobs.asmx?op=UploadData
        """

        wsid=''
        password=''
        if configfile is None:
            configfile = "CasJobs.config"
        logger.info("Reading config file: %s"%configfile)
        lines = open(configfile,'r').readlines()
        for line in lines:
            k,v = line.strip().split('=')
            if k == 'wsid': wsid = v
            if k == 'password': password = v

        logger.info("Attempting to drop table: %s"%table)
        self.drop(table)
     
        SOAP_TEMPLATE = """
        <soap12:Envelope xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" 
                         xmlns:xsd="http://www.w3.org/2001/XMLSchema" 
                         xmlns:soap12="http://www.w3.org/2003/05/soap-envelope">
          <soap12:Body>
            <UploadData xmlns="http://Services.Cas.jhu.edu">
              <wsid>%s</wsid>
              <pw>%s</pw>
              <tableName>%s</tableName>
              <data>%s</data>
              <tableExists>%s</tableExists>
            </UploadData>
          </soap12:Body>
        </soap12:Envelope>
        """
     
        logger.info("Writing array...")
        s = StringIO.StringIO()
        np.savetxt(s,array,delimiter=',',fmt="%.10g")
        tb_data = ''
        if fields is not None: 
            tb_data += ','.join(f for f in fields)+'\n'
        tb_data += s.getvalue()
     
        message = SOAP_TEMPLATE % (wsid, password, table, tb_data, "false")
        
        #construct and send the header
        webservice = httplib.HTTP("skyserver.sdss3.org")
        webservice.putrequest("POST", "/casjobs/services/jobs.asmx")
        webservice.putheader("Host", "skyserver.sdss3.org")
        webservice.putheader("Content-type", "text/xml; charset=\"UTF-8\"")
        webservice.putheader("Content-length", "%d" % len(message))
        webservice.endheaders()
        logger.info("Sending SOAP POST message...")
        webservice.send(message)
         
        # get the response
        statuscode, statusmessage, header = webservice.getreply()
        print "Response: ", statuscode, statusmessage
        print "headers: ", header
        res = webservice.getfile().read()
        print res


    def run(self,pixfile=None,outdir=None):
        self.load_pixels(pixfile)
        for pixel in self.pixels:
            self.download(pixel,outdir)

    def inFootprint(self, ra, dec):
        basename = self.basename + '_coverage'
        sqlname = basename + '.sql'
        dbname = basename + '_output'
        task = basename
        outfile = basename+".fits"

        # Upload the (ra,dec) coordinates to casjobs
        table = "ra_dec"
        self.upload(np.array([ra,dec]).T, ['ra','dec'],table=table)

        # Query the database for the footprint
        query = open(sqlname,'w')
        query.write('SELECT dbo.fInFootprintEq(t.ra, t.dec, 0)\n')
        query.write('INTO MyDB.%s\n' % (dbname))
        query.write('FROM MyDB.%s AS t'%table)
        query.close()

        self.query(self.release,task,sqlname)
        url = self.extract(dbname)
        
        if (url is not None):
            self.wget(url,outfile)

        self.drop(table)
        self.drop(dbname)


    def footprint(self,nside):
        """
        Download the survey footprint for HEALpix pixels.
        """
        import healpy
        import ugali.utils.projector
        if nside > 2**9: raise Exception("Overflow error: nside must be <=2**9")
        pix = numpy.arange(healpy.nside2npix(nside),dtype='int')
        footprint = numpy.zeros(healpy.nside2npix(nside),dtype='bool')
        ra,dec = ugali.utils.projector.pixToAng(nside,pix)
        table_name = 'Pix%i'%nside
        self.upload(np.array([pix,ra,dec]).T, ['pix','ra','dec'], name=table_name)
        radius = healpy.nside2resol(nside_superpix,arcmin=True)

        query="""
        SELECT t.pix, dbo.fInFootprintEq(t.ra, t.dec, %g)
        FROM %s AS t
        """%(radius, table_name)

class DESDatabase(Database):
    ####################################
    ###### !!! NOT IMPLEMENTED !!! #####
    ####################################

    def __init__(self,release='SVA1_GOLD'):
        super(DESDatabase,self).__init__()
        self.release = release.lower()
        self.basename = "des_%s_photometry"%self.release

    def _setup_desdbi(self):
        # Function here to setup trivialAccess client...
        # This should work but it doesn't
        import despydb.desdbi
        import pyfits

    def generate_query(self, ra_min,ra_max,dec_min,dec_max,filename,db):
        # Preliminary and untested
        table = 'Y1A1_COADD_OBJECTS@DESSCI'

        outfile = open(filename,"w")
        outfile.write('SELECT s.COADD_OBJECTS_ID, s.RA, s.DEC, \n')
        outfile.write('s.MAG_PSF_G, s.MAGERR_PSF_G, \n')
        outfile.write('s.MAG_PSF_R, s.MAGERR_PSF_R, \n')
        outfile.write('s.MAG_PSF_I, s.MAGERR_PSF_I  \n')
        outfile.write('FROM %s s \n'%(table))
        outfile.write('WHERE s.MODEST_CLASS = 2 \n')
        outfile.write('AND s.RA > %.7f AND s.RA < %.7f \n' % (ra_min,ra_max))
        outfile.write('AND s.DEC > %.7f AND s.DEC < %.7f;' % (dec_min,dec_max))
        outfile.write('> %s \n' % db)
        outfile.close()

    ### def query(self,dbase,task,query):
    ###     import despydb.desdbi
    ###     logger.info("Running query...")
    ###  
    ###     desfile = os.path.join(os.getenv('HOME'),'.desservices.ini')
    ###     section = 'db-dessci'
    ###     dbi = despydb.desdbi.DesDbi(desfile,section)
    ###     cursor = dbi.cursor()
    ###     cursor.execute(open(query,'r').read())
    ###  
    ###     header = [d[0] for d in cursor.description]
    ###     data = cursor.fetchall()
    ###        
    ###     logger.info("Found %i rows."%len(data))
    ###     if not len(header) or not len(data): return None
    ###     array = np.rec.array(data,names=header)
    ###     return array

    def query(self,dbase,task,query):
        logger.info("Running query...")
        cmd = "easyaccess -l %s" % (query)
        logger.info(cmd)
        return subprocess.call(cmd,shell=True)
        #ret = subprocess.check_output(cmd,shell=True,stderr=subprocess.STDOUT) 
        #if 'ERROR:' in ret:
        #    raise subprocess.CalledProcessError(1,cmd,ret)
        #return ret

    def download(self, pixel, outdir=None, force=False):
        import pyfits 

        if outdir is None: outdir = './'
        else:              mkdir(outdir)
        sqldir = mkdir(os.path.join(outdir,'sql'))
        self._setup_desdbi()

        basename = self.basename + "_%04d"%pixel['name']
        sqlname = os.path.join(sqldir,basename+'.sql')
        taskname = basename
        outfile = os.path.join(outdir,basename+".fits")
        # ADW: There should be a 'force' option here
        if os.path.exists(outfile) and not force:
            logger.warning("Found %s; skipping..."%(outfile))
            return

        logger.info("\nDownloading pixel: %(name)i (ra=%(ra_min)g:%(ra_max)g,dec=%(dec_min)g:%(dec_max)g)"%(pixel))
        logger.info("Working on "+sqlname)
         
        self.generate_query(pixel['ra_min'],pixel['ra_max'],pixel['dec_min'],pixel['dec_max'],sqlname,outfile)
        ret = self.query(self.release,taskname,sqlname)
        if ret != 0:
            msg = "Download failed to complete."
            raise Exception(msg)
        return outfile

    def run(self,pixfile=None,outdir=None):
        self.load_pixels(pixfile)
        for pixel in self.pixels:
            self.download(pixel,outdir)

class Y1A1Database(DESDatabase):
    def __init__(self):
        release='Y1A1'
        super(Y1A1Database,self).__init__(release=release)
        self.release = release.lower()
        self.basename = "des_%s_photometry"%self.release
        print self.basename

    def generate_query(self, ra_min,ra_max,dec_min,dec_max,filename,db):
        # Preliminary and untested
        table = 'Y1A1_COADD_OBJECTS@DESSCI'

        select = 'ABS(s.WAVG_SPREAD_MODEL_I) < 0.004\n'
        select += 'AND s.FLAGS_G < 4 and s.FLAGS_R < 4 and s.FLAGS_I < 4\n'
        select += 'AND s.MAG_AUTO_G between 0 and 30 and s.MAG_AUTO_R between 0 and 30 and s.MAG_AUTO_I between 0 and 30\n'
        select += 'AND s.MAGERR_AUTO_G < 1 and s.MAGERR_AUTO_R < 1 and s.MAGERR_AUTO_I < 1'
        #select = '1 = 1'

        outfile = open(filename,"w")
        outfile.write('-- %s \n'%(self.__class__.__name__))
        outfile.write('SELECT s.COADD_OBJECTS_ID, s.RA, s.DEC, \n')
        outfile.write('s.MAG_PSF_G-XCORR_SFD98_G as MAGSFD_PSF_G, s.MAGERR_PSF_G, \n')
        outfile.write('s.MAG_PSF_R-XCORR_SFD98_R as MAGSFD_PSF_R, s.MAGERR_PSF_R, \n')
        outfile.write('s.MAG_PSF_I-XCORR_SFD98_I as MAGSFD_PSF_I, s.MAGERR_PSF_I, \n')
        outfile.write('s.MAG_AUTO_G-XCORR_SFD98_G as MAGSFD_AUTO_G, s.MAGERR_AUTO_G, \n')
        outfile.write('s.MAG_AUTO_R-XCORR_SFD98_R as MAGSFD_AUTO_R, s.MAGERR_AUTO_R, \n')
        outfile.write('s.MAG_AUTO_I-XCORR_SFD98_I as MAGSFD_AUTO_I, s.MAGERR_AUTO_I, \n')
        outfile.write('s.WAVG_SPREAD_MODEL_I, s.WAVG_SPREADERR_MODEL_I, \n')
        outfile.write('s.SPREAD_MODEL_I, s.SPREADERR_MODEL_I  \n')
        outfile.write('FROM %s s \n'%(table))
        outfile.write('WHERE %s \n'%(select))
        outfile.write('AND s.RA > %.7f AND s.RA < %.7f \n' % (ra_min,ra_max))
        outfile.write('AND s.DEC > %.7f AND s.DEC < %.7f; \n' % (dec_min,dec_max))
        outfile.write(' > %s \n' % (db))
        outfile.close()


class Y2NDatabase(DESDatabase):
    def __init__(self):
        release='Y2N'
        super(Y2NDatabase,self).__init__(release=release)

    def generate_query(self, ra_min,ra_max,dec_min,dec_max,filename,db):
        # Preliminary and untested
        table = 'kadrlica.Y2N_UNIQUE_OBJECTS@DESOPER'

        select = 'ABS(s.WAVG_SPREAD_MODEL_R) < 0.003\n'
        select += 'AND s.WAVG_MAG_PSF_G between 0 and 30 and s.WAVG_MAG_PSF_R between 0 and 30\n'
        select += 'AND s.MAGERR_AUTO_G < 1 and s.MAGERR_AUTO_R < 1'
        #select = '1 = 1'

        outfile = open(filename,"w")
        outfile.write('-- %s \n'%(self.__class__.__name__))
        outfile.write('SELECT s.CATALOG_ID, s.RA, s.DEC, \n')
        outfile.write('s.WAVG_MAG_PSF_G, s.MAGERR_PSF_G, \n')
        outfile.write('s.WAVG_MAG_PSF_R, s.MAGERR_PSF_R, \n')
        outfile.write('s.WAVG_MAG_AUTO_G, s.MAGERR_AUTO_G, \n')
        outfile.write('s.WAVG_MAG_AUTO_R, s.MAGERR_AUTO_R, \n')
        outfile.write('s.WAVG_SPREAD_MODEL_R, s.WAVG_SPREADERR_MODEL_R, \n')
        outfile.write('s.SPREAD_MODEL_R, s.SPREADERR_MODEL_R \n')
        outfile.write('FROM %s s \n'%(table))
        outfile.write('WHERE %s; \n'%(select))
        outfile.write(' > %s \n'%(db))
        outfile.close()

class Y2Q1Database(DESDatabase):
    def __init__(self):
        release='Y2Q1'
        super(Y2Q1Database,self).__init__(release=release)

    def generate_query(self, ra_min,ra_max,dec_min,dec_max,filename,db):
        # Preliminary and untested
        table = 'kadrlica.Y2Q1_OBJECTS_V1@DESOPER'

        select = 'ABS(s.WAVG_SPREAD_MODEL_R) < 0.003 + s.SPREADERR_MODEL_R \n'
        select += 'AND s.WAVG_MAG_PSF_G between 0 and 30 and s.WAVG_MAG_PSF_R between 0 and 30\n'
        select += 'AND s.WAVG_MAGERR_PSF_G < 1 and s.WAVG_MAGERR_PSF_R < 1'

        outfile = open(filename,"w")
        outfile.write('-- %s \n'%(self.__class__.__name__))
        outfile.write('SELECT s.CATALOG_ID, s.RA, s.DEC, \n')
        outfile.write('s.MAG_PSF_G, s.MAGERR_PSF_G, \n')
        outfile.write('s.WAVG_MAG_PSF_G, s.WAVG_MAGERR_PSF_G, \n')
        outfile.write('s.MAG_PSF_R, s.MAGERR_PSF_R, \n')
        outfile.write('s.WAVG_MAG_PSF_R, s.WAVG_MAGERR_PSF_R, \n')
        outfile.write('s.SPREAD_MODEL_R, s.SPREADERR_MODEL_R, \n')
        outfile.write('s.WAVG_SPREAD_MODEL_R, s.CLASS_STAR_R \n')
        outfile.write('FROM %s s \n'%(table))
        outfile.write('WHERE %s; \n'%(select))
        outfile.write(' > %s \n'%(db))
        outfile.close()

if __name__ == "__main__":
    import ugali.utils.parser
    description = "Download data set."
    parser = ugali.utils.parser.Parser(description=description)
    parser.add_config()
    parser.add_debug()
    parser.add_verbose()
    parser.add_argument('pixfile',metavar='pixels.dat',default=None,
                        nargs='?',help='Input pixel file.')
    opts = parser.parse_args()

    from ugali.utils.config import Config
    config = Config(opts.config)

    survey = config['data']['survey'].lower()
    release = config['data']['release'].lower()
    outdir = config['data']['dirname']

    db = databaseFactory(survey,release)
    db.run(pixfile=opts.pixfile,outdir=outdir)



"""
Documentation.
"""

import scipy.special
import numpy

############################################################

def gammalnStirling(z):
    """
    Uses Stirling's approximation for the log-gamma function suitable for large arguments.
    """
    return (0.5 * (numpy.log(2. * numpy.pi) - numpy.log(z))) \
           + (z * (numpy.log(z + (1. / ((12. * z) - (1. / (10. * z))))) - 1.))

############################################################

def confidenceInterval(n, k, alpha = 0.68, errorbar=False):
    """
    Given n tests and k successes, return efficiency and confidence interval.
    """
    e = float(k) / float(n)
    bins = 1000001
    dx = 1. / bins

    efficiency = numpy.linspace(0, 1, bins)

    # MODIFIED FOR LARGE NUMBERS
    if n + 2 > 1000:
        a = gammalnStirling(n + 2)
    else:
        a = scipy.special.gammaln(n + 2)
    if k + 1 > 1000:
        b = gammalnStirling(k + 1)
    else:
        b = scipy.special.gammaln(k + 1)
    if n - k + 1 > 1000:
        c = gammalnStirling(n - k + 1)
    else:
        c = scipy.special.gammaln(n - k + 1)

    if k == 0:
        p = numpy.concatenate([[numpy.exp(a - b - c)],
                               numpy.exp(a - b - c + (k * numpy.log(efficiency[1: -1])) + (n - k) * numpy.log(1. - efficiency[1: -1])),
                               [0.]])
    elif k == n:
        p = numpy.concatenate([[0.],
                               numpy.exp(a - b - c + (k * numpy.log(efficiency[1: -1])) + (n - k) * numpy.log(1. - efficiency[1: -1])),
                               [numpy.exp(a - b - c)]])
    else:
        p = numpy.concatenate([[0.],
                               numpy.exp(a - b - c + (k * numpy.log(efficiency[1: -1])) + (n - k) * numpy.log(1. - efficiency[1: -1])),
                               [0.]])

    i = numpy.argsort(p)[::-1]
    p_i = numpy.take(p, i)

    s = i[numpy.cumsum(p_i * dx) < alpha]

    low = min(numpy.min(s) * dx, e)
    high = max(numpy.max(s) * dx, e)

    if not errorbar:
        return e, [low, high]
    else:
        return e, [e - low, high - e]

############################################################

bayesianInterval = confidenceInterval

############################################################

def binomialInterval(n, k, alpha = 0.68):
    """
    Given n tests and k successes, return efficiency and confidence interval.
    """
    e = float(k)/n
    delta_e = 1/float(n) * numpy.sqrt(e * (1 - e) * float(n)) * alpha/0.68
    return e, [e - delta_e, e + delta_e]

############################################################



#!/usr/bin/env python
import os
import yaml
from collections import OrderedDict as odict
import numpy as np
import copy

from ugali.analysis.model import Model, Parameter
from ugali.analysis.kernel import factory as kernelFactory
from ugali.analysis.isochrone import factory as isochroneFactory

class Richness(Model):
    """Dummy model to hold the richness, which is not directly connected
    to either the spatial or color information and doesn't require a
    sync when updated.
    """
    _params = odict([
        ('richness', Parameter(1000.0, [0.0,  np.inf])),
    ])

# Just to be consistent with other factories
def richnessFactory(type='Richness',**kwargs):
    return Richness(**kwargs)

class Source(object):
    """ 
    A source model builds the components on the spatial and spectral model.

    ADW: We could probably be smarter and get this to inherit from Model...
    """
    _defaults = odict([
            ('richness' ,dict(name='Richness')),
            ('kernel'   ,dict(name='Plummer')),
            ('isochrone',dict(name='Padova')),
            ])

    def __init__(self,name=None, **kwargs):
        self.set_model('richness',self.createRichness())
        self.set_model('kernel',self.createKernel())
        self.set_model('isochrone',self.createIsochrone())

        # Toggle for tracking which models need to be synched
        self._sync = odict([(k,True) for k in self.models.keys()])

        self.name = name
        self.set_params(**kwargs)

    def __str__(self):
        ret = "%s : Name=%s"%(self.__class__.__name__,self.name)
        for key,model in self.models.items():
            ret += "\n  %s Model (sync=%s):\n"%(key.capitalize(),self._sync[key])
            ret += model.__str__(indent=4)
        return ret

    def __getattr__(self, name):
        #for key,model in self.models.items():
        #    if name in model.params:
        #        return getattr(model, name)
        #for key,model in self.models.items():
        #    try:
        #        return model.getp(name).value
        #    except KeyError:
        #        continue
        ## Raises AttributeError
        #return object.__getattribute__(self,name)
        try:
            return self.getp(name)
        except AttributeError:
            return object.__getattribute__(self,name)

    def __setattr__(self, name, value):
        #for key,model in self.models.items():
        #    if name in model.params:
        #        self._sync[key] = True
        #        return setattr(model, name, value)
        #for key,model in self.models.items():
        #    try:
        #        ret = model.setp(name, value)
        #        self._sync[key] = True
        #        return ret
        #    except KeyError:
        #        continue
        ## Raises AttributeError?
        #return object.__setattr__(self, name, value)
        try:
            return self.setp(name,value)
        except AttributeError:
            return object.__setattr__(self, name, value)

    def setp(self, name, *args, **kwargs):
        for key,model in self.models.items():
            try:
                ret = model.setp(name, *args, **kwargs)
                self._sync[key] = True
                return ret
            except KeyError:
                continue
        raise AttributeError

    def getp(self, name):
        for key,model in self.models.items():
            try:
                return model.getp(name).value
            except KeyError:
                continue
        raise AttributeError

    @property
    def params(self):
        # DANGEROUS: Altering properties directly doesn't call model._cache
        params = odict([])
        for key,model in self.models.items():
            params.update(model.params)
        return params


    def load(self,srcmdl,section=None):
        if isinstance(srcmdl,basestring): 
            params = yaml.load(open(srcmdl))
        else:
            params = copy.deepcopy(srcmdl)
        if section is not None: 
            params = params[section]

        fill = False
        if params.get('name'):
            self.name = params.get('name')
        if params.get('richness'):
            richness  = self.createRichness(**params['richness'])
            self.set_model('richness',richness)
            fill = True
        if params.get('kernel'):
            kernel    = self.createKernel(**params['kernel'])
            self.set_model('kernel',kernel)
            fill = True
        if params.get('isochrone'):
            isochrone = self.createIsochrone(**params['isochrone'])
            self.set_model('isochrone',isochrone)
            fill = True

        if not fill:
            msg = "Didn't load any source parameters."
            raise Exception(msg)

    def dump(self):
        return yaml.dump(self.todict())

    def todict(self):
        ret = odict()
        if self.name is not None: ret['name'] = self.name
        for name,model in self.models.items():
            ret[name] = model.todict()
        return ret
        
    # ADW: Should be class methods
    @staticmethod
    def createRichness(**kwargs):
        for k,v in copy.deepcopy(Source._defaults['richness']).items():
            kwargs.setdefault(k,v)
        return richnessFactory(**kwargs)

    @staticmethod
    def createKernel(**kwargs):
        for k,v in copy.deepcopy(Source._defaults['kernel']).items():
            kwargs.setdefault(k,v)
        return kernelFactory(**kwargs)

    @staticmethod
    def createIsochrone(**kwargs):
        for k,v in copy.deepcopy(Source._defaults['isochrone']).items():
            kwargs.setdefault(k,v)
        return isochroneFactory(**kwargs)
    
    # Various derived properties
    @property
    def kernel(self):
        return self.models['kernel']

    @property
    def isochrone(self):
        return self.models['isochrone']

    def set_model(self, name, model):
        """ Set a model """
        if not hasattr(self,'models'):
            object.__setattr__(self, 'models',odict())
        self.models[name] = model

    def set_kernel(self,kernel): 
        self.set_model('kernel',kernel)

    def set_isochrone(self,isochrone): 
        self.set_model('isochrone',isochrone)

    def set_params(self,**kwargs):
        """ Set the parameter values """
        for key,value in kwargs.items():
            setattr(self,key,value)

    def get_params(self):
        """ Get an odict of the parameter names and values """
        return odict([(key,param.value) for key,param in self.params.items()])

    def get_free_params(self):
        """ Get an odict of free parameter names and values """
        return odict([(key,param.value) for key,param in self.params.items() if param.free])

    def set_free_params(self, names):
        for name in self.get_params().keys():
            self.setp(name,free=False)
        for name in np.array(names,ndmin=1):
            self.setp(name,free=True)
    
    def get_sync(self,model):
        return self._sync.get(model)

    def reset_sync(self):
        for k in self._sync.keys(): self._sync[k]=False

    def read(self,filename):
        pass

    def write(self,filename,force=False):
        if os.path.exists(filename) and not force:
            raise Exception("Found %s..."%filename)
        out = open(filename,'w')
        out.write(self.dump())
        out.close()

    def set_stellar_mass(self, stellar_mass):
        """ Set the richness to match an input stellar mass. """
        self.richness = stellar_mass/self.isochrone.stellar_mass()

    #@property
    def stellar_mass(self):
        # ADW: I think it makes more sense for this to be
        #return self.richness * self.isochrone.stellarMass()
        return self.isochrone.stellarMass()

    #@property
    def stellar_luminosity(self):
        # ADW: I think it makes more sense for this to be
        #return self.richness * self.isochrone.stellarLuminosity()
        return self.isochrone.stellarLuminosity()

    #@property
    def absolute_magnitude(self, richness):
        #ADW:  Should take self.richness
        return self.isochrone.absolute_magnitude(richness)

if __name__ == "__main__":
    import argparse
    description = "python script"
    parser = argparse.ArgumentParser(description=description)
    opts = parser.parse_args()



#!/usr/bin/env python
"""
Download Padova isochrones from:
http://stev.oapd.inaf.it/cgi-bin/cmd

Adapted from ezpadova by Morgan Fouesneau:
https://github.com/mfouesneau/ezpadova

"""
import os
from urllib import urlencode
from urllib2 import urlopen
from StringIO import StringIO
import zlib
import re
import subprocess
from collections import OrderedDict as odict

import numpy as np
from ugali.utils.logger import logger
from ugali.utils.shell import mkdir
from ugali.analysis.isochrone import PadovaIsochrone,OldPadovaIsochrone

# Commented options may need to be restored for older
# version/isochrones.
defaults_27 =  {#'binary_frac': 0.3,
                #'binary_kind': 1,
                #'binary_mrinf': 0.7,
                #'binary_mrsup': 1,
                'cmd_version': 2.7,
                'dust_source': 'nodust',
                'dust_sourceC': 'nodustC',
                'dust_sourceM': 'nodustM',
                'eta_reimers': 0.2,
                #'extinction_av': 0,
                #'icm_lim': 4,
                'imf_file': 'tab_imf/imf_chabrier_lognormal.dat',
                'isoc_age': 1e9,
                'isoc_age0': 12.7e9,
                'isoc_dlage': 0.05,
                'isoc_dz': 0.0001,
                'isoc_kind': 'parsec_CAF09_v1.2S',
                'isoc_lage0': 6.602,   #Minimum allowed age                 
                'isoc_lage1': 10.1303, #Maximum allowed age                 
                'isoc_val': 0,               
                'isoc_z0': 0.0001,     #Minimum allowed metallicity         
                'isoc_z1': 0.03,       #Maximum allowed metallicity   
                'isoc_zeta': 0.0002,
                'isoc_zeta0': 0.0002,
                'kind_cspecmag': 'aringer09',
                'kind_dust': 0,
                'kind_interp': 1,
                'kind_mag': 2,
                'kind_postagb': -1,
                'kind_pulsecycle': 0,
                #'kind_tpagb': 0,
                #'lf_deltamag': 0.2,
                #'lf_maginf': 20,
                #'lf_magsup': -20,
                #'mag_lim': 26,
                #'mag_res': 0.1,
                'output_evstage': 1,
                'output_gzip': 0,
                'output_kind': 0,
                'photsys_file': 'tab_mag_odfnew/tab_mag_decam.dat',
                #'photsys_version': 'yang',
                'submit_form': 'Submit'}

class Padova(object):
    defaults = dict(defaults_27)

    params2filename = PadovaIsochrone.params2filename
    filename2params = PadovaIsochrone.filename2params

    def __init__(self,survey='des',**kwargs):
        self.survey=survey

    def create_grid(self,abins,zbins):
        arange = np.linspace(abins[0],abins[1],abins[2]+1)
        zrange = np.logspace(np.log10(zbins[0]),np.log10(zbins[1]),zbins[2]+1)
        aa,zz = np.meshgrid(arange,zrange)
        return aa.flatten(),zz.flatten()

    def run(self,grid=None,outdir=None,force=False):
        if grid is None:
            aa,zz = self.create_grid()
        else:
            aa,zz = grid
        for a,z in zip(aa,zz):
            try: 
                self.download(a,z,outdir,force)
            except RuntimeError, msg:
                logger.warning(msg)

    def download(self,age,metallicity,outdir=None,force=False):
        """
        Check valid parameter range and download isochrones from:
        http://stev.oapd.inaf.it/cgi-bin/cmd
        """
        epsilon = 1e-4
        lage = np.log10(age*1e9)
        lage_min,lage_max = self.defaults['isoc_lage0'],self.defaults['isoc_lage1']
        if not (lage_min-epsilon < lage <lage_max+epsilon):
            msg = 'Age outside of valid range: %g [%g < log(age) < %g]'%(lage,lage_min,lage_max)
            raise RuntimeError(msg)

        z_min,z_max = self.defaults['isoc_z0'],self.defaults['isoc_z1']
        if not (z_min <= metallicity <= z_max):
            msg = 'Metallicity outside of valid range: %g [%g < z < %g]'%(metallicity,z_min,z_max)
            raise RuntimeError(msg)

        survey=self.survey.lower()
        if survey=='des':
            photsys_file='tab_mag_odfnew/tab_mag_decam.dat'
        elif survey=='sdss':
            photsys_file='tab_mag_odfnew/tab_mag_sloan.dat'
        else:
            msg = 'Unrecognized survey: %s'%survey
            raise RuntimeError(msg)

        if outdir is None: outdir = './'
        mkdir(outdir)

        basename = self.params2filename(age,metallicity)
        outfile = os.path.join(outdir,basename)
            
        if os.path.exists(outfile) and not force:
            logger.warning("Found %s; skipping..."%(outfile))
            return

        logger.info("Downloading isochrone: %s (age=%.2fGyr, metallicity=%g)"%(basename,age,metallicity))

        d = dict(self.defaults)
        d['photsys_file'] = photsys_file
        d['isoc_age']     = age * 1e9
        d['isoc_zeta']    = metallicity

        server = 'http://stev.oapd.inaf.it'
        url = server + '/cgi-bin/cmd_%s'%d['cmd_version']
        logger.info("Accessing %s..."%url)

        q = urlencode(d)
        logger.debug(url+'?'+q)
        c = urlopen(url, q).read()
        aa = re.compile('output\d+')
        fname = aa.findall(c)
        if len(fname) > 0:
            out = '{0}/~lgirardi/tmp/{1}.dat'.format(server, fname[0])
            cmd = 'wget %s -O %s'%(out,outfile)
            logger.debug(cmd)
            stdout = subprocess.check_output(cmd,shell=True,stderr=subprocess.STDOUT)
            logger.debug(stdout)
        else:
            #print(c)
            raise RuntimeError('Server Response is incorrect')

class OldPadova(Padova):
    defaults = dict(defaults_27)
    defaults['isoc_kind'] = 'gi10a'

    params2filename = OldPadovaIsochrone.params2filename
    filename2params = OldPadovaIsochrone.filename2params

class Girardi2002(Padova):
    defaults = dict(defaults_27)
    defaults['isoc_kind'] = 'gi2000'

class Marigo2008(Padova):
    defaults = dict(defaults_27)
    defaults['isoc_kind'] = 'ma08'

class Girardi2010a(Padova):
    defaults = dict(defaults_27)
    defaults['isoc_kind'] = 'gi10a'

class Girardi2010b(Padova):
    defaults = dict(defaults_27)
    defaults['isoc_kind'] = 'gi10b'

class Bressan2012(Padova):
    defaults = dict(defaults_27)
    defaults['isoc_kind'] = 'parsec_CAF09_v1.2S'


if __name__ == "__main__":
    import ugali.utils.parser
    description = "Download isochrones"
    parser = ugali.utils.parser.Parser(description=description)
    parser.add_config()
    parser.add_debug()
    parser.add_verbose()
    parser.add_force()
    parser.add_argument('-a','--age',default=None,type=float)
    parser.add_argument('-z','--metallicity',default=None,type=float)
    opts = parser.parse_args()

    from ugali.utils.config import Config
    config = Config(opts.config)

    #outdir = '/u/ki/kadrlica/des/isochrones/v1'
    #outdir = '/u/ki/kadrlica/sdss/isochrones/v2'
    #outdir = '/u/ki/kadrlica/des/isochrones/v2'
    #outdir = '/u/ki/kadrlica/des/isochrones/v3'
    #outdir = '/u/ki/kadrlica/des/isochrones/v4'
    outdir = '/u/ki/kadrlica/des/isochrones/v5'
    #outdir = '/u/ki/kadrlica/des/isochrones/v6'
    survey = config['data']['survey']

    #p = Padova(survey=survey)
    p = Girardi2010a(survey=survey)
    #p = Girardi2010b(survey=survey)
    #p = Marigo2008(survey=survey)
    p = Girardi2002(survey=survey)

    # Binning from config
    #abins = config['binning']['age']
    #zbins = config['binning']['z']
    #abins = np.arange(1,13.6,0.1)
    #zbins = np.arange(1e-4,1e-3,1e-5)
    abins = np.arange(10,14,1)
    zbins = np.arange(1e-4,2e-4,1e-4)
    #abins = np.arange(10,10.1,0.1)
    #zbins = np.arange(1e-4,1.1e-4,1e-5)


    if opts.age is not None: abins = [opts.age]
    if opts.metallicity is not None: zbins = [opts.metallicity]
    grid = [g.flatten() for g in np.meshgrid(abins,zbins)]
        
    #grid = p.create_grid(abins,zbins)
    #grid = OldPadovaIsochrone.create_grid()
    #grid = np.meshgrid((10**np.arange(9.9,10.15,0.05))/1e9,np.array([0.12,0.15,0.19,0.24,0.30,0.38,0.48,0.6])*1e-3)
    #cut = (grid[0] > 0.5)
    #grid = (grid[0][cut],grid[1][cut])
    #grid = np.meshgrid(np.arange(1,13.5),np.arange(1e-4,1e-3,5e-5)

    print "Ages:"
    print np.unique(grid[0])
    print "Metallicities:"
    print np.unique(grid[1])
    p.run(grid=grid,outdir=outdir,force=opts.force)



#!/usr/bin/env python
import time

def speedtest(func, *args, **kwargs):
    """ Test the speed of a function. """
    n = 100
    start = time.time()
    for i in range(n): func(*args,**kwargs)
    end = time.time()
    return (end-start)/n

if __name__ == "__main__":
    from optparse import OptionParser
    usage = "Usage: %prog  [options] input"
    description = "python script"
    parser = OptionParser(usage=usage,description=description)
    (opts, args) = parser.parse_args()
